import numbers

import numpy as np

from .backend import get_backend
from .validation import check_random_state


def compute_lipschitz_constants(Xs, kernelize="XTX", random_state=None):
    """Compute Lipschitz constants of gradients of linear regression problems.

    Find the largest eigenvalue of X^TX for several X, using power iteration.

    Parameters
    ----------
    Xs : array of shape (n_kernels, n_samples, n_features) or \
            (n_kernels, n_samples, n_samples)
        Multiple linear features or kernels.
    kernelize : str in {"XTX", "XXT", "X"}
        Whether to consider X^TX, XX^T, or directly X.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.

    Returns
    -------
    lipschitz : array of shape (n_kernels)
        Lipschitz constants.
    """
    backend = get_backend()

    if kernelize == "XXT":
        XTs = backend.transpose(Xs, (0, 2, 1))
        kernels = backend.matmul(Xs, XTs)
        del XTs
    elif kernelize == "XTX":
        XTs = backend.transpose(Xs, (0, 2, 1))
        kernels = backend.matmul(XTs, Xs)
        del XTs
    elif kernelize == "X":
        kernels = Xs
    else:
        raise ValueError("Unknown parameter kernelize=%r" % (kernelize, ))

    # check the random state
    random_generator = check_random_state(random_state)
    ys = random_generator.randn(*(kernels.shape[:2] + (1, )))

    ys = backend.asarray_like(ys, Xs)
    for i in range(10):
        ys /= backend.norm(ys, axis=1, keepdims=True) + 1e-16
        ys = backend.matmul(kernels, ys)
    evs = backend.norm(ys, axis=1)[:, 0]
    return evs


def assert_array_almost_equal(x, y, decimal=6, err_msg='', verbose=True):
    """Test array equality, casting all arrays to numpy."""
    backend = get_backend()
    x = backend.to_numpy(x)
    y = backend.to_numpy(y)
    return np.testing.assert_array_almost_equal(x, y, decimal=decimal,
                                                err_msg=err_msg,
                                                verbose=verbose)


def generate_multikernel_dataset(n_kernels=4, n_targets=500,
                                 n_samples_train=1000, n_samples_test=400,
                                 noise=0.1, kernel_weights=None,
                                 n_features_list=None, random_state=None):
    """Utility to generate datasets for the gallery of examples.

    Parameters
    ----------
    n_kernels : int
        Number of kernels.
    n_targets : int
        Number of targets.
    n_samples_train : int
        Number of samples in the training set.
    n_samples_test : int
        Number of sample in the testing set.
    noise : float > 0
        Scale of the Gaussian white noise added to the targets.
    kernel_weights : array of shape (n_targets, n_kernels) or None
        Kernel weights used in the prediction of the targets.
        If None, generate random kernel weights from a Dirichlet distribution.
    n_features_list : list of int of length (n_kernels, ) or None
        Number of features in each kernel. If None, use 1000 features for each.
    random_state : int, or None
        Random generator seed use to generate the true kernel weights.

    Returns
    -------
    X_train : array of shape (n_samples_train, n_features)
        Training features.
    X_test : array of shape (n_samples_test, n_features)
        Testing features.
    Y_train : array of shape (n_samples_train, n_targets)
        Training targets.
    Y_test : array of shape (n_samples_test, n_targets)
        Testing targets.
    kernel_weights : array of shape (n_targets, n_kernels)
        Kernel weights in the prediction of the targets.
    n_features_list : list of int of length (n_kernels, )
        Number of features in each kernel.
    """
    from .kernel_ridge import generate_dirichlet_samples
    backend = get_backend()

    # Create a few kernel weights if not given.
    if kernel_weights is None:
        kernel_weights = generate_dirichlet_samples(n_targets, n_kernels,
                                                    concentration=[.3],
                                                    random_state=random_state)
        kernel_weights = backend.to_numpy(kernel_weights)

    if n_features_list is None:
        n_features_list = np.full(n_kernels, fill_value=1000)

    rng = check_random_state(random_state)

    # Then, generate a random dataset, using the arbitrary scalings.
    Xs_train, Xs_test = [], []
    Y_train, Y_test = None, None
    for ii in range(n_kernels):
        n_features = n_features_list[ii]

        X_train = rng.randn(n_samples_train, n_features)
        X_test = rng.randn(n_samples_test, n_features)
        X_train -= X_train.mean(0)
        X_test -= X_test.mean(0)
        Xs_train.append(X_train)
        Xs_test.append(X_test)

        weights = rng.randn(n_features, n_targets) / n_features
        weights *= kernel_weights[:, ii] ** 0.5

        if ii == 0:
            Y_train = X_train @ weights
            Y_test = X_test @ weights
        else:
            Y_train += X_train @ weights
            Y_test += X_test @ weights

    std = Y_train.std(0)[None]
    Y_train /= std
    Y_test /= std

    Y_train += rng.randn(n_samples_train, n_targets) * noise
    Y_test += rng.randn(n_samples_test, n_targets) * noise
    Y_train -= Y_train.mean(0)
    Y_test -= Y_test.mean(0)

    # Concatenate the feature spaces.
    X_train = backend.asarray(np.concatenate(Xs_train, 1), dtype="float32")
    X_test = backend.asarray(np.concatenate(Xs_test, 1), dtype="float32")
    Y_train = backend.asarray(Y_train, dtype="float32")
    Y_test = backend.asarray(Y_test, dtype="float32")
    kernel_weights = backend.asarray(kernel_weights, dtype="float32")

    return X_train, X_test, Y_train, Y_test, kernel_weights, n_features_list


def _batch_or_skip(array, batch, axis):
    """Apply a batch on given axis, or skip if the dimension is equal to 1."""
    skip = (array is None or isinstance(array, numbers.Number)
            or array.ndim == 0 or array.shape[axis] == 1)  # noqa
    if skip:
        return array
    else:
        # Not general but works with slices in `batch`.
        if axis == 0:
            return array[batch]
        elif axis == 1:
            return array[:, batch]
        else:
            raise NotImplementedError()



from ._utils import ALL_BACKENDS
from ._utils import CURRENT_BACKEND
from ._utils import set_backend
from ._utils import get_backend
from ._utils import force_cpu_backend

__all__ = [
    "ALL_BACKENDS",
    "CURRENT_BACKEND",
    "set_backend",
    "get_backend",
    "force_cpu_backend",
]



import warnings
import itertools

import numpy as np

from ..utils import compute_lipschitz_constants
from ..progress_bar import bar, ProgressBar
from ..backend import get_backend
from ..scoring import r2_score
from ..validation import check_cv


def solve_sparse_group_lasso_cv(X, Y, groups=None, l21_regs=[0.05],
                                l1_regs=[0.05], cv=5, max_iter=300, tol=1e-4,
                                momentum=True, n_targets_batch=None,
                                progress_bar=True):
    """Solves the sparse group Lasso, selecting hyperparameters over
    cross-validation.

    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        Input features.
    Y : array of shape (n_samples, n_targets)
        Target data.
    groups : array of shape (n_features, ) or None
        Encoding of the group of each feature. If None, all features are
        gathered in one group, and the problem is equivalent to the Lasso.
    l21_regs : array of shape (n_l21_regs, )
        All the Group Lasso regularization parameter tested.
    l1_regs : array of shape (n_l1_regs, )
        All the Lasso regularization parameter tested.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    max_iter : int > 0
        Maximum number of iterations.
    tol : float > 0
        Tolerance of the stopping criterion.
    momentum : bool
        If True, use acceleration in the projected gradient descent.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.

    Returns
    -------
    coef : array of shape (n_features, n_targets)
        Coefficient of the linear model. Always on CPU.
    best_l21_reg : array of shape (n_targets, )
        Best hyperparameter per target.
    best_l1_reg : array of shape (n_targets, )
        Best hyperparameter per target.
    all_cv_scores : array of shape (n_l21_regs * n_l1_regs, n_targets)
        Cross-validation scores of all tested hyperparameters.
    """
    backend = get_backend()

    cv = check_cv(cv, Y)
    n_splits = cv.get_n_splits()

    X, Y, l21_regs, l1_regs = backend.check_arrays(X, Y, l21_regs, l1_regs)
    n_samples, n_features = X.shape
    n_samples, n_targets = Y.shape
    if n_targets_batch is None:
        n_targets_batch = n_targets

    l21_regs = backend.asarray_like(backend.atleast_1d(l21_regs), ref=Y)
    l1_regs = backend.asarray_like(backend.atleast_1d(l1_regs), ref=Y)

    # sort in decreasing order, since it is faster with warm starting
    decreasing_l21_regs = backend.flip(backend.sort(l21_regs), 0)
    if not backend.all(decreasing_l21_regs == l21_regs):
        l21_regs = decreasing_l21_regs
        warnings.warn("l21_regs has been reordered in decreasing order.")

    # sort in decreasing order, since it is faster with warm starting
    decreasing_l1_regs = backend.flip(backend.sort(l1_regs), 0)
    if not backend.all(decreasing_l1_regs == l1_regs):
        l1_regs = decreasing_l1_regs
        warnings.warn("l1_regs has been reordered in decreasing order.")

    n_l21_regs = l21_regs.shape[0]
    n_l1_regs = l1_regs.shape[0]
    n_batches = len(range(0, n_targets, n_targets_batch))

    if progress_bar:
        progress_bar = ProgressBar(
            f"grid search cv over {n_l21_regs * n_l1_regs} parameters",
            max_value=n_l21_regs * n_l1_regs * n_splits * n_batches)

    coef = None
    all_cv_scores = backend.zeros_like(
        X, shape=(n_l21_regs * n_l1_regs, n_targets))
    best_l21_reg = backend.zeros_like(X, shape=(n_targets, ))
    best_l1_reg = backend.zeros_like(X, shape=(n_targets, ))

    for kk, (train, val) in enumerate(cv.split(Y)):
        if hasattr(Y, "device"):
            val = backend.asarray(val, device=Y.device)
            train = backend.asarray(train, device=Y.device)

        lipschitz_train = compute_lipschitz_constants(X[train][None],
                                                      kernelize="XTX")[0]

        for start in range(0, n_targets, n_targets_batch):
            batch = slice(start, start + n_targets_batch)

            # reset coef, to avoid leaking info from split to split
            coef = None
            for ii, (l21_reg,
                     l1_reg) in enumerate(itertools.product(l21_regs,
                                                            l1_regs)):

                coef = solve_sparse_group_lasso(
                    X[train], Y[train][:, batch], groups=groups,
                    l21_reg=l21_reg, l1_reg=l1_reg, max_iter=max_iter, tol=tol,
                    momentum=momentum, initial_coef=coef,
                    lipschitz=lipschitz_train, n_targets_batch=n_targets_batch,
                    progress_bar=False)

                Y_val_pred = X[val] @ coef
                scores = r2_score(Y[val][:, batch], Y_val_pred)

                all_cv_scores[ii, batch] += scores

                if progress_bar:
                    progress_bar.update_with_increment_value(1)

    # select best hyperparameter configuration
    all_cv_scores /= n_splits
    argmax = backend.argmax(all_cv_scores, 0)
    config = backend.asarray(
        list(
            itertools.product(backend.to_numpy(l21_regs),
                              backend.to_numpy(l1_regs))))
    best_config = config[argmax]
    best_l21_reg = best_config[:, 0]
    best_l1_reg = best_config[:, 1]

    # refit
    coef_cpu = backend.zeros_like(X, shape=(n_features, n_targets),
                                  device="cpu")
    for start in range(0, n_targets, n_targets_batch):
        batch = slice(start, start + n_targets_batch)
        coef = solve_sparse_group_lasso(
            X, Y[:, batch], groups=groups, l21_reg=best_l21_reg[batch],
            l1_reg=best_l1_reg[batch], max_iter=max_iter, tol=tol,
            momentum=momentum, initial_coef=None, lipschitz=None,
            n_targets_batch=None, progress_bar=False)
        coef_cpu[:, batch] = backend.to_cpu(coef)
    return coef_cpu, best_l21_reg, best_l1_reg, all_cv_scores


def solve_sparse_group_lasso(X, Y, groups=None, l21_reg=0.05, l1_reg=0.05,
                             max_iter=300, tol=1e-4, momentum=True,
                             initial_coef=None, lipschitz=None,
                             n_targets_batch=None, progress_bar=True,
                             debug=False):
    """Solves the sparse group Lasso.

    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        Input features.
    Y : array of shape (n_samples, n_targets)
        Target data.
    groups : array of shape (n_features, ) or None
        Encoding of the group of each feature. If None, all features are
        gathered in one group, and the problem is equivalent to the Lasso.
    l21_reg : float, or array of shape (n_targets, )
        Regularization parameter for the group Lasso.
    l1_reg : float, or array of shape (n_targets, )
        Regularization parameter for the Lasso.
    max_iter : int > 0
        Maximum number of iterations.
    tol : float > 0
        Tolerance of the stopping criterion.
    momentum : bool
        If True, use acceleration in the projected gradient descent.
    initial_coef : None, or array of shape (n_features, n_targets)
        Initial value for the projected gradient descent.
    lipschitz : float or None
        Lipschitz constant of the gradient. If None, it will be estimated from
        X using power iterations.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.
    debug : bool
        If True, compute the objective function at each iteration, and return
        the list of results (of the last batch) along with the final
        coefficients.

    Returns
    -------
    coef : array of shape (n_features, n_targets)
        Coefficient of the linear model.
    """
    backend = get_backend()

    n_samples, n_features = X.shape
    n_samples, n_targets = Y.shape
    if n_targets_batch is None:
        n_targets_batch = n_targets

    X, Y = backend.check_arrays(X, Y)

    if lipschitz is None:
        lipschitz = compute_lipschitz_constants(X[None], kernelize="XTX")[0]
    else:
        lipschitz = backend.asarray_like(lipschitz, ref=X)

    if groups is None:
        groups = backend.zeros((n_features))
    groups = backend.asarray(groups)[:]
    groups = [groups == u for u in backend.unique(groups) if u >= 0]

    if initial_coef is None:
        coef = backend.zeros_like(X, shape=(n_features, n_targets))
    else:
        coef = backend.asarray_like(initial_coef, ref=X)

    l1_reg = l1_reg * n_samples  # as in scikit-learn
    l21_reg = l21_reg * n_samples

    l1_reg = backend.asarray_like(backend.atleast_1d(l1_reg), ref=Y)
    if l1_reg.shape[0] == 1:
        l1_reg = backend.ones_like(Y, shape=(n_targets, )) * l1_reg[0]
    l21_reg = backend.asarray_like(backend.atleast_1d(l21_reg), ref=Y)
    if l21_reg.shape[0] == 1:
        l21_reg = backend.ones_like(Y, shape=(n_targets, )) * l21_reg[0]

    use_l1_reg = any(l1_reg > 0)
    use_l21_reg = any(l21_reg > 0)

    use_it = progress_bar and n_targets > n_targets_batch
    for bb, start in enumerate(
            bar(range(0, n_targets, n_targets_batch), title="Group Lasso",
                use_it=use_it)):
        batch = slice(start, start + n_targets_batch)

        def loss(ww):
            """Objective function"""
            error_sq = 0.5 * ((X @ ww - Y[:, batch]) ** 2).sum(0)

            if use_l1_reg:
                error_sq += l1_reg[batch] * backend.abs(ww).sum(0)

            if use_l21_reg:
                for group in groups:
                    error_sq += l21_reg[batch] * backend.sqrt(
                        (ww[group] ** 2).sum(0))

            return error_sq

        def grad(ww, mask=slice(0, None)):
            """Gradient of the objective function"""
            return X.T @ (X @ ww - Y[:, batch][:, mask])

        def prox(ww, mask=slice(0, None)):
            """Proximal operator for the projected gradient descent"""
            if use_l1_reg:
                ww = _l1_prox(ww, l1_reg[batch][mask] / lipschitz)
            if use_l21_reg:
                ww = _l21_prox(ww, l21_reg[batch][mask] / lipschitz, groups)
            return ww

        tmp = _proximal_gradient_descent(
            loss, grad, prox, step_size=1. / lipschitz, x0=coef[:, batch],
            max_iter=max_iter, momentum=momentum, tol=tol,
            progress_bar=progress_bar and n_targets <= n_targets_batch,
            debug=debug)
        if debug:
            tmp, losses = tmp
        coef[:, batch] = tmp

    if debug:
        return coef, losses
    else:
        return coef


def _l1_prox(ww, reg):
    """The proximal operator for reg * ||ww||_1."""
    backend = get_backend()
    return backend.sign(ww) * backend.clip(backend.abs(ww) - reg, 0, None)


def _sqrt_l2_prox(ww, reg):
    """The proximal operator for reg * ||ww||_2 (not squared)."""
    backend = get_backend()

    norm_ww = backend.norm(ww, axis=0)
    mask = norm_ww == 0

    ww[:, mask] = 0
    ww[:, ~mask] = backend.clip(1 - reg[~mask] / norm_ww[~mask], 0,
                                None)[None] * ww[:, ~mask]
    return ww


def _l21_prox(ww, reg, groups):
    """The proximal operator for reg * ||ww||_21 for all groups"""
    backend = get_backend()

    ww = backend.copy(ww)
    for group in groups:
        ww[group, :] = _sqrt_l2_prox(ww[group, :], reg)

    return ww


###############################################################################
# fista algorithm


def _proximal_gradient_descent(f_loss, f_grad, f_prox, step_size, x0, max_iter,
                               momentum=True, tol=1e-7, progress_bar=True,
                               debug=False):
    """Proximal Gradient Descent (PGD) and Accelerated PDG.

    This reduces to ISTA and FISTA when the loss function is the l2 loss and
    the proximal operator is the soft-thresholding.

    Parameters
    ----------
    f_loss : callable
        Objective function. Only used when debug = True.
    f_grad : callable
        Gradient of the objective function.
    f_prox : callable
        Proximal operator.
    step_size : float
        Step size of each update.
    x0 : array of shape (n_features, n_targets)
        Initial point of the optimization.
    max_iter : int
        Maximum number of iterations.
    momentum : bool
        If True, use FISTA instead of ISTA.
    tol : float or None
        Tolerance for the stopping criterion.
    progress_bar : bool
        If True, display a progress bar over iterations.
    debug : bool
        If True, compute the objective function at each iteration, and return
        the list of results along with the final point x_hat.

    Returns
    -------
    x_hat : array of shape (n_features, n_targets)
        The final point after optimization.
    losses : array of shape (n_iter_ + 1, n_targets)
        Returned only if debug = True. Objective function at each iteration.
    """
    backend = get_backend()

    if debug:
        losses = [f_loss(x0)]

    t_new = 1.0
    x_hat = backend.copy(x0)

    # auxiliary variables
    x_old = backend.copy(x_hat)
    x_aux = backend.copy(x_hat)
    diff = backend.zeros_like(x_hat)

    # not converged targets
    mask = backend.ones_like(x0, dtype=backend.bool, shape=(x0.shape[1]))

    if progress_bar:
        progress_bar = ProgressBar(title='fista', max_value=max_iter)
    for ii in range(max_iter):
        # apply gradient and prox, from x_aux
        x_old[:] = x_hat[:, mask]
        x_hat[:, mask] = x_aux - step_size * f_grad(x_aux, mask)
        x_hat[:, mask] = f_prox(x_hat[:, mask], mask)
        diff = x_hat[:, mask] - x_old

        # apply momentum, from x_hat
        x_aux[:] = x_hat[:, mask]
        if momentum:
            t_old = t_new
            t_new = (1 + np.sqrt(1 + 4 * t_old * t_old)) / 2
            x_aux += (t_old - 1.) / t_new * diff

        if debug:
            losses.append(f_loss(x_hat))

        if tol is not None:
            criterion = (backend.norm(diff, axis=0) /
                         (backend.norm(x_hat[:, mask], axis=0) + 1e-16))
            just_converged = criterion <= tol

            if backend.any(just_converged):
                x_aux = x_aux[:, ~just_converged]
                x_old = x_old[:, ~just_converged]
                mask[backend.copy(mask)] = ~just_converged

                if backend.all(just_converged):
                    break

        if progress_bar:
            progress_bar.update_with_increment_value(1)
    else:
        warnings.warn("FISTA did not converge.", RuntimeWarning)

    if progress_bar:
        progress_bar.close()

    if debug:
        return x_hat, backend.stack(losses)

    return x_hat



import warnings

from .backend import get_backend


def l2_neg_loss(y_true, y_pred):
    """L2 negative loss, computed for multiple predictions.

    Parameters
    ----------
    y_true : array or Tensor of shape (n_samples, n_targets)
        Ground truth.
    y_pred : array or Tensor of shape (n_predictions, n_samples, n_targets) or\
            (n_samples, n_targets)
        Predictions.

    Returns
    -------
    l2 : array of shape (n_predictions, n_targets) or (n_targets, )
        L2 negative losses.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    # broadcasting if y_pred has more dimensions
    axis = 1 if len(y_pred.shape) > len(y_true.shape) else 0
    n_samples = y_true.shape[0]

    error = ((y_true - y_pred) ** 2).sum(axis)
    l2 = -error / n_samples

    return l2


def r2_score(y_true, y_pred):
    """R2 score, computed for multiple predictions (e.g. multiple alphas).

    Parameters
    ----------
    y_true : array or Tensor of shape (n_samples, n_targets)
        Ground truth.
    y_pred : array or Tensor of shape (n_predictions, n_samples, n_targets) or\
            (n_samples, n_targets)
        Predictions.

    Returns
    -------
    r2 : array of shape (n_predictions, n_targets) or (n_targets, )
        R2 scores.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    # broadcasting if y_pred has more dimensions
    axis = 1 if len(y_pred.shape) > len(y_true.shape) else 0

    error = ((y_true - y_pred) ** 2).sum(axis)
    var = ((y_true - y_true.mean(0)) ** 2).sum(0)
    r2 = 1. - error / var
    if axis == 0:
        r2[var == 0] = 0
    else:
        r2[:, var == 0] = 0

    return r2


def correlation_score(y_true, y_pred):
    """Correlation score, computed for multiple predictions.

    Parameters
    ----------
    y_true : array or Tensor of shape (n_samples, n_targets)
        Ground truth.
    y_pred : array or Tensor of shape (n_predictions, n_samples, n_targets) or\
            (n_samples, n_targets)
        Predictions.

    Returns
    -------
    correlations : array of shape (n_predictions, n_targets) or (n_targets, )
        Correlation scores.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    # broadcasting if y_pred has more dimensions
    axis = 1 if len(y_pred.shape) > len(y_true.shape) else 0

    product = _zscore(y_true, axis=0) * _zscore(y_pred, axis=axis)
    correlations = product.mean(axis)

    return correlations


def r2_score_split(y_true, y_pred, include_correlation=True):
    """Split the R2 score into individual components using the product measure.

    When estimating a linear joint model, the predictions of each feature space
    are summed::

        Yhat_joint = Yhat_A + Yhat_B + ... + Yhat_Z

    The joint model R2 can be computed as::

        R2_joint = R2(Yhat_joint, Y)

    This function estimates the contribution of each feature space to the joint
    model R2 such that::

        R2_joint = R2_A + R2_B + ... + R2_Z

    Mathematically, this is achieved by taking into account the correlations
    between predictions (i.e. Yhat_A*Yhat_B,..., Yhat_A*Yhat_Z). The function
    can also returns an estimate that ignores these correlations.

    This function differs from r2_score_split_svd in the method used to
    decompose the variance. The function r2_score_split is based on the product
    measure method, while the function r2_score_split_svd is based on the
    relative weights method.

    This function assumes that y_true is zero-mean over samples.

    Parameters
    ----------
    y_true : array of shape (n_samples, n_targets)
        Observed data. Has to be zero-mean over samples.
    y_pred : array of shape (n_kernels, n_samples, n_targets) or \
            (n_samples, n_targets)
        Predictions.
    include_correlation : bool
        Whether to include correlation between feature spaces.
        If True, individual feature space R2 sum is equivalent to the joint
        model R2 (i.e. from `y_pred.sum(0)`).

    Returns
    -------
    r2 : array (n_kernels, n_targets) or (n_targets, )
        Individual feature space R2 scores.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    if backend.any(backend.abs(y_true.mean(0)) > 1e-6):
        warnings.warn(
            'y_true has to be zero-mean over samples to compute '
            'the split r2 scores.', UserWarning)

    sst = (y_true ** 2).sum(0)

    no_split = y_pred.ndim == 2
    if no_split:
        y_pred = y_pred[None]

    n_kernels, n_samples, n_targets = y_pred.shape
    r2 = backend.zeros_like(y_pred[:, 0, :])

    for fsi in range(n_kernels):
        if include_correlation:
            asst = (y_pred[fsi] * y_pred).sum(0)
        else:
            asst = (y_pred[fsi] ** 2)

        inter = y_true * y_pred[fsi]
        r2[fsi, :] = ((2 * inter - asst) / sst).sum(0)
        del inter, asst

    if no_split:
        r2 = r2[0]

    return r2


def r2_score_split_svd(y_true, y_pred):
    """Split the R2 score into individual components using relative weights.

    When estimating a linear joint model, the predictions of each feature space
    are summed::

        Yhat_joint = Yhat_A + Yhat_B + ... + Yhat_Z

    The joint model R2 can be computed as::

        R2_joint = R2(Yhat_joint, Y)

    This function estimates the contribution of each feature space to the joint
    model R2 such that::

        R2_joint = R2_A + R2_B + ... + R2_Z

    This function differs from r2_score_split in the method used to decompose
    the variance. The function r2_score_split is based on the product measure
    method, while the function r2_score_split_svd is based on the relative
    weights method.

    This function assumes that y_true is zero-mean over samples.

    Parameters
    ----------
    y_true : array of shape (n_samples, n_targets)
        Observed data. Has to be zero-mean over samples.
    y_pred : array of shape (n_kernels, n_samples, n_targets) or \
            (n_samples, n_targets)
        Predictions.

    Returns
    -------
    r2 : array (n_kernels, n_targets) or (n_targets, )
        Individual feature space R2 scores.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    if backend.any(backend.abs(y_true.mean(0)) > 1e-6):
        warnings.warn(
            'y_true has to be zero-mean over samples to compute '
            'the split r2 scores.', UserWarning)

    no_split = y_pred.ndim == 2
    if no_split:
        y_pred = y_pred[None]

    # compute the R2 score on the sum of sub-predictions
    y_pred_sum = y_pred.sum(0)
    full_r2 = r2_score(y_true, y_pred_sum)

    # normalize the prediction sum
    norms_pred_sum = backend.norm(y_pred_sum, axis=0, keepdims=True)
    norms_pred_sum[norms_pred_sum == 0] = 1
    y_pred_sum = y_pred_sum / norms_pred_sum
    y_pred = y_pred / norms_pred_sum

    norms_pred = backend.norm(y_pred, axis=1)
    norms_pred[norms_pred == 0] = 1

    # Compute the singular value decomposition. It is done on non-normalized
    # y_pred to preserve the scaling of sub-predictions.
    U, s, Vt = backend.svd(backend.transpose(y_pred, (2, 1, 0)),
                           full_matrices=False)
    V = backend.transpose(Vt, (0, 2, 1))
    Ut = backend.transpose(U, (0, 2, 1))

    # regression of y_pred over V.U^T
    VsVt = V @ (s[:, :, None] * Vt)
    # regression of y_pred_sum over V.U^T
    beta = (V @ Ut) @ y_pred_sum.T[:, :, None]

    # Aggregates VsVt and beta to compute the relative weights. It uses a
    # scaling by norms_pred because the SVD was computed on non-normalized
    # y_pred.
    r2 = backend.matmul(VsVt ** 2 / norms_pred.T[:, None, :] ** 2,
                        beta ** 2)[:, :, 0].T

    # fix rounding errors
    r2 /= r2.sum(0)[None, :]
    # scale by the total R2
    r2 *= full_r2[None, :]

    if no_split:
        r2 = r2[0]

    return r2


def correlation_score_split(y_true, y_pred):
    """Split the correlation score into individual components.

    When estimating a linear joint model, the predictions of each feature space
    are summed::

        Yhat_joint = Yhat_A + Yhat_B + ... + Yhat_Z

    The joint model correlation score r can be computed as::

        r_joint = r(Y, Yhat_joint)

    This function estimates the contribution of each feature space to the joint
    model correlation score r such that::

        r_joint = r_A + r_B + ... + r_Z

    Parameters
    ----------
    y_true : array or Tensor of shape (n_samples, n_targets)
        Ground truth.
    y_pred : array or Tensor of shape (n_predictions, n_samples, n_targets) or\
            (n_samples, n_targets)
        Predictions.

    Returns
    -------
    correlations : array of shape (n_predictions, n_targets) or (n_targets, )
        Contributions of each individual feature space to the joint correlation
        score.
    """
    backend = get_backend()
    y_true, y_pred = backend.check_arrays(y_true, y_pred)
    y_pred = _check_finite(y_pred)

    # broadcasting if y_pred has more dimensions
    axis = 1 if len(y_pred.shape) > len(y_true.shape) else 0
    # demean to take the covariance
    y_true = y_true - y_true.mean(0, keepdims=True)
    y_pred = y_pred - y_pred.mean(axis, keepdims=True)
    y_true_std = backend.std_float64(y_true, 0, demean=False, keepdims=False)
    y_true_std[y_true_std == 0] = 1
    split = y_pred.ndim == 3
    if split:
        y_true = y_true[None]
        y_pred_sum = y_pred.sum(0, keepdims=True)
        y_pred_std = backend.std_float64(y_pred_sum, axis, demean=True,
                                         keepdims=False)
    else:
        y_pred_std = backend.std_float64(y_pred, axis, demean=True,
                                         keepdims=True)
    y_pred_std[y_pred_std == 0] = 1
    correlations = (y_true * y_pred).mean(axis) / (y_true_std * y_pred_std)
    if not split:
        correlations = correlations[0]
    return correlations


###############################################################################


def _check_finite(y_pred):
    backend = get_backend()

    is_nan = backend.isnan(y_pred)
    if backend.any(is_nan):
        warnings.warn('nan in y_pred.')
        y_pred[is_nan] = 0

    isinf = backend.isinf(y_pred)
    if backend.any(isinf):
        warnings.warn('inf in y_pred.')
        y_pred[isinf] = 0

    return y_pred


def _zscore(X, axis):
    backend = get_backend()

    Xz = X - X.mean(axis, keepdims=True)
    std = backend.std_float64(Xz, axis, demean=False, keepdims=True)
    std[std == 0] = 1
    Xz /= std
    return Xz



"""Adapt functions from scikit-learn to use different backends."""

import itertools
import math
from functools import partial

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from ..backend import force_cpu_backend, get_backend
from ..validation import _get_string_dtype, check_array, issparse


def check_pairwise_arrays(X, Y, precomputed=False, dtype=None,
                          accept_sparse='csr', force_all_finite=True):
    """ Set X and Y appropriately and checks inputs

    If Y is None, it is set as a pointer to X (i.e. not a copy).
    If Y is given, this does not happen.
    All distance metrics should use this function first to assert that the
    given parameters are correct and safe to use.

    Specifically, this function first ensures that both X and Y are arrays,
    then checks that they are at least two dimensional while ensuring that
    their elements are floats (or dtype if provided). Finally, the function
    checks that the size of the second dimension of the two arrays is equal, or
    the equivalent check for a precomputed distance matrix.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape (n_samples_X, n_features)

    Y : {array-like, sparse matrix}, shape (n_samples_Y, n_features)

    precomputed : bool
        True if X is to be treated as precomputed distances to the samples in
        Y.

    dtype : string, type, list of types or None (default=None)
        Data type required for X and Y. If None, the dtype will be an
        appropriate float type selected by _return_float_dtype.

    accept_sparse : string, boolean or list/tuple of strings
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc. If the input is sparse but not in the allowed format,
        it will be converted to the first listed format. True allows the input
        to be any format. False means that a sparse matrix input will
        raise an error.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in array. The
        possibilities are:

        - True: Force all values of array to be finite.
        - False: accept both np.inf and np.nan in array.
        - 'allow-nan': accept only np.nan values in array. Values cannot
          be infinite.

    Returns
    -------
    X : {array-like, sparse matrix}, shape (n_samples_X, n_features)
        An array equal to X.

    Y : {array-like, sparse matrix}, shape (n_samples_Y, n_features)
        An array equal to Y if Y was not None.
        If Y was None, Y will be a pointer to X.

    """
    if dtype is None:
        dtype = _return_float_dtype(X, Y)

    if Y is X or Y is None:
        X = Y = check_array(X, accept_sparse=accept_sparse, dtype=dtype,
                            force_all_finite=force_all_finite)
    else:
        X = check_array(X, accept_sparse=accept_sparse, dtype=dtype,
                        force_all_finite=force_all_finite)
        Y = check_array(Y, accept_sparse=accept_sparse, dtype=dtype,
                        force_all_finite=force_all_finite)

    if precomputed:
        pass
    elif X.shape[1] != Y.shape[1]:
        raise ValueError("Incompatible dimension for X and Y matrices: "
                         "X.shape[1] == %d while Y.shape[1] == %d" %
                         (X.shape[1], Y.shape[1]))

    return X, Y


def _return_float_dtype(X, Y):
    """
    1. If dtype of X and Y is float32, then dtype float32 is returned.
    2. Else dtype float64 is returned.
    """
    X_dtype = _get_string_dtype(X)

    if Y is None:
        Y_dtype = X_dtype
    else:
        Y_dtype = _get_string_dtype(Y)

    if X_dtype == Y_dtype == "float32":
        dtype = "float32"
    else:
        dtype = "float64"

    return dtype


def _row_norms(X, squared=False):
    """Row-wise (squared) Euclidean norm of X.

    Equivalent to np.sqrt((X * X).sum(axis=1)), but also supports sparse
    matrices and does not create an X.shape-sized temporary.

    Performs no input validation.

    Parameters
    ----------
    X : array_like
        The input array
    squared : bool, optional (default = False)
        If True, return squared norms.

    Returns
    -------
    array_like
        The row-wise (squared) Euclidean norm of X.
    """
    backend = get_backend()

    if issparse(X):
        import scipy.sparse
        if not isinstance(X, scipy.sparse.csr_matrix):
            X = scipy.sparse.csr_matrix(X)
        from sklearn.utils.sparsefuncs_fast import csr_row_norms
        norms = csr_row_norms(X)
    else:
        norms = backend.einsum('ij,ij->i', X, X)

    if not squared:
        backend.sqrt(norms, out=norms)
    return norms


def _normalize(X):
    """L2 normalizer with copy."""
    if issparse(X):
        X = X.copy()
        from sklearn.utils.sparsefuncs_fast import inplace_csr_row_normalize_l2
        inplace_csr_row_normalize_l2(X)
    else:
        norms = _row_norms(X)
        norms[norms == 0.0] = 1.0
        X = X / norms[:, None]
    return X


###############################################################################


def linear_kernel(X, Y=None):
    """
    Compute the linear kernel between X and Y.

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.

    Returns
    -------
    K : array of shape (n_samples_X, n_samples_Y)
        Computed kernel.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)

    # Convert to backend arrays before matrix operations to handle transpose properly
    if issparse(X):
        X = X.toarray()
    if issparse(Y):
        Y = Y.toarray()
    X = backend.asarray(X)
    Y = backend.asarray(Y)
    
    K = backend.matmul(X, backend.transpose(Y))
    return K


def polynomial_kernel(X, Y=None, degree=3, gamma=None, coef0=1):
    """
    Compute the polynomial kernel between X and Y::

        K(X, Y) = (gamma <X, Y> + coef0)^degree

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.
    degree : int, default 3
        Degree of the polynomial.
    gamma : float, default None
        if None, defaults to 1.0 / n_features
    coef0 : float, default 1
        Intercept.

    Returns
    -------
    K : array of shape (n_samples_X, n_samples_Y)
        Computed kernel.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)
    if gamma is None:
        gamma = 1.0 / X.shape[1]

    # Convert to backend arrays before matrix operations to handle transpose properly
    if issparse(X):
        X = X.toarray()
    if issparse(Y):
        Y = Y.toarray()
    X = backend.asarray(X)
    Y = backend.asarray(Y)

    K = backend.matmul(X, backend.transpose(Y))
    K *= gamma
    K += coef0
    K = backend.power(K, degree, out=K)
    return K


def sigmoid_kernel(X, Y=None, gamma=None, coef0=1):
    """
    Compute the sigmoid kernel between X and Y::

        K(X, Y) = tanh(gamma <X, Y> + coef0)

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.
    gamma : float, default None
        If None, defaults to 1.0 / n_features
    coef0 : float, default 1
        Intercept.

    Returns
    -------
    K : array of shape (n_samples_X, n_samples_Y)
        Computed kernel.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)
    if gamma is None:
        gamma = 1.0 / X.shape[1]

    K = X @ Y.T
    if issparse(K):
        K = K.toarray()
    K = backend.asarray(K)
    K *= gamma
    K += coef0
    K = backend.tanh(K, out=K)
    return K


def rbf_kernel(X, Y=None, gamma=None):
    """
    Compute the rbf (gaussian) kernel between X and Y::

        K(x, y) = exp(-gamma ||x-y||^2)

    for each pair of rows x in X and y in Y.

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.
    gamma : float, default None
        If None, defaults to 1.0 / n_features

    Returns
    -------
    K : array of shape (n_samples_X, n_samples_Y)
        Computed kernel.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)
    if gamma is None:
        gamma = 1.0 / X.shape[1]

    K = euclidean_distances(X, Y, squared=True)
    K = backend.asarray(K)
    K *= -gamma
    K = backend.exp(K, out=K)
    return K


def cosine_similarity_kernel(X, Y=None):
    """Compute cosine similarity between samples in X and Y.

    Cosine similarity, or the cosine kernel, computes similarity as the
    normalized dot product of X and Y:

        K(X, Y) = <X, Y> / (||X||*||Y||)

    On L2-normalized data, this function is equivalent to linear_kernel.

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.

    Returns
    -------
    K : array of shape (n_samples_X, n_samples_Y)
        Computed kernel.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)

    X_normalized = _normalize(X)
    if X is Y:
        Y_normalized = X_normalized
    else:
        Y_normalized = _normalize(Y)

    K = X_normalized @ Y_normalized.T

    if issparse(K):
        K = K.toarray()
    K = backend.asarray(K)

    return K


PAIRWISE_KERNEL_FUNCTIONS = {
    'linear': linear_kernel,
    'polynomial': polynomial_kernel,
    'poly': polynomial_kernel,
    'rbf': rbf_kernel,
    'sigmoid': sigmoid_kernel,
    'cosine': cosine_similarity_kernel,
}

###############################################################################


def euclidean_distances(X, Y=None, squared=False):
    """
    Considering the rows of X (and Y=X) as vectors, compute the
    distance matrix between each pair of vectors.

    Parameters
    ----------
    X : array of shape (n_samples_X, n_features)
        Train or test features.
    Y : array of shape (n_samples_Y, n_features)
        Train features.
    squared : boolean, optional
        Whether to return squared distances.

    Returns
    -------
    distances : array, shape (n_samples_X, n_samples_Y)
        Computed distances.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y)

    distances = _euclidean_distances_upcast(X, Y)
    if not squared:
        distances = backend.sqrt(distances, out=distances)

    return distances


def _euclidean_distances_upcast(X, Y, batch_size=None):
    """Euclidean distances between X and Y.

    X and Y are upcast to float64 by chunks, which size is chosen to limit
    memory increase by approximately 10% (at least 10MiB).
    """
    backend = get_backend()

    n_samples_X = X.shape[0]
    n_samples_Y = Y.shape[0]
    n_features = X.shape[1]

    if issparse(X):
        distances = backend.zeros(shape=(n_samples_X, n_samples_Y),
                                  dtype=_get_string_dtype(X))
    else:
        distances = backend.zeros_like(X, shape=(n_samples_X, n_samples_Y))

    if batch_size is None:
        X_size = backend.prod(backend.asarray(X.shape))
        Y_size = backend.prod(backend.asarray(Y.shape))
        x_density = X.nnz / X_size if issparse(X) else 1
        y_density = Y.nnz / Y_size if issparse(Y) else 1

        # Allow 10% more memory than X, Y and the distance matrix take (at
        # least 10MiB)
        maxmem = max(
            ((x_density * n_samples_X + y_density * n_samples_Y) * n_features +
             (x_density * n_samples_X * y_density * n_samples_Y)) / 10.,
            10 * 2 ** 17)

        # The increase amount of memory in 8-byte blocks is:
        # - x_density * batch_size * n_features (copy of chunk of X)
        # - y_density * batch_size * n_features (copy of chunk of Y)
        # - batch_size * batch_size (chunk of distance matrix)
        # Hence x² + (xd+yd)kx = M, where x=batch_size, k=n_features, M=maxmem
        #                                 xd=x_density and yd=y_density
        tmp = (x_density + y_density) * n_features
        batch_size = (-tmp + math.sqrt(tmp ** 2 + 4 * maxmem)) / 2
        batch_size = max(int(batch_size), 1)

    for x_start in range(0, n_samples_X, batch_size):
        x_batch = slice(x_start, x_start + batch_size)

        if issparse(X):
            X_chunk = X[x_batch].astype("float64", copy=False)
        else:
            X_chunk = backend.asarray(X[x_batch], dtype=backend.float64)
        XX_chunk = _row_norms(X_chunk, squared=True)[:, None]

        for y_start in range(0, n_samples_Y, batch_size):
            y_batch = slice(y_start, y_start + batch_size)

            if X is Y and y_batch < x_batch:
                # when X is Y the distance matrix is symmetric so we only need
                # to compute half of it.
                d = distances[y_batch, x_batch].T
            else:
                if issparse(Y):
                    Y_chunk = Y[y_batch].astype("float64", copy=False)
                else:
                    Y_chunk = backend.asarray(Y[y_batch],
                                              dtype=backend.float64)
                YY_chunk = _row_norms(Y_chunk, squared=True)[None, :]

                d = -2 * X_chunk @ Y_chunk.T
                if issparse(d):
                    d = d.toarray()
                d += XX_chunk
                d += YY_chunk

            if issparse(X):
                distances[x_batch, y_batch] = backend.asarray(d)
            else:
                distances[x_batch, y_batch] = backend.asarray_like(d, X)

    return distances


def _pairwise_callable(X, Y, metric, force_all_finite=True, **params):
    """Handle the callable case for pairwise_{distances,kernels}.
    """
    backend = get_backend()
    X, Y = check_pairwise_arrays(X, Y, force_all_finite=force_all_finite)
    out = backend.zeros_like(X, shape=(X.shape[0], Y.shape[0]), dtype='float')

    if X is Y:
        # Only calculate metric for upper triangle
        iterator = itertools.combinations(range(X.shape[0]), 2)
        for i, j in iterator:
            out[i, j] = metric(X[i], Y[j], **params)

        # Make symmetric
        out = out + out.T

        # Calculate diagonal
        # NB: nonzero diagonals are allowed for both metrics and kernels
        for i in range(X.shape[0]):
            x = X[i]
            out[i, i] = metric(x, x, **params)

    else:
        # Calculate all cells
        iterator = itertools.product(range(X.shape[0]), range(Y.shape[0]))
        for i, j in iterator:
            out[i, j] = metric(X[i], Y[j], **params)

    return out


def pairwise_kernels(X, Y=None, metric="linear", n_jobs=None, **params):
    """Compute the kernel between arrays X and optional array Y.

    This method takes either a vector array or a kernel matrix, and returns
    a kernel matrix. If the input is a vector array, the kernels are
    computed. If the input is a kernel matrix, it is returned instead.

    This method provides a safe way to take a kernel matrix as input, while
    preserving compatibility with many other algorithms that take a vector
    array.

    If Y is given (default is None), then the returned matrix is the pairwise
    kernel between the arrays from both X and Y.

    Valid values for metric are:
        ['additive_chi2', 'chi2', 'linear', 'poly', 'polynomial', 'rbf',
        'laplacian', 'sigmoid', 'cosine']

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    X : array of shape (n_samples_X, n_samples_X) if metric == "precomputed", \
             (n_samples_X, n_features) otherwise
        Array of pairwise kernels between samples, or a feature array.

    Y : array of shape (n_samples_Y, n_features)
        A second feature array only if X metric != "precomputed".

    metric : string, or callable
        The metric to use when calculating kernel between instances in a
        feature array. If metric is a string, it must be one of the metrics
        in PAIRWISE_KERNEL_FUNCTIONS.
        If metric is "precomputed", X is assumed to be a kernel matrix.
        Alternatively, if metric is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded. The callable
        should take two rows from X as input and return the corresponding
        kernel value as a single number. This means that callables from
        :mod:`sklearn.metrics.pairwise` are not allowed, as they operate on
        matrices, not single samples. Use the string identifying the kernel
        instead.

    **params : optional keyword parameters
        Any further parameters are passed directly to the kernel function.

    Returns
    -------
    K : array [n_samples_X, n_samples_X] or [n_samples_X, n_samples_Y]
        A kernel matrix K such that K_{i, j} is the kernel between the
        ith and jth vectors of the given matrix X, if Y is None.
        If Y is not None, then K_{i, j} is the kernel between the ith array
        from X and the jth array from Y.

    Notes
    -----
    If metric is 'precomputed', Y is ignored and X is returned.
    """
    if metric == "precomputed":
        X, _ = check_pairwise_arrays(X, Y, precomputed=True)
        return X

    elif metric in PAIRWISE_KERNEL_FUNCTIONS:
        func = PAIRWISE_KERNEL_FUNCTIONS[metric]
    elif callable(metric):
        func = partial(_pairwise_callable, metric=metric, **params)
    else:
        raise ValueError("Unknown metric=%r." % metric)

    return func(X, Y, **params)


class KernelCenterer(TransformerMixin, BaseEstimator):
    """Center a kernel matrix.

    Adapt sklearn.preprocessing.KernelCenterer to use other backends.

    Let K(x, z) be a kernel defined by phi(x)^T phi(z), where phi is a
    function mapping x to a Hilbert space. KernelCenterer centers (i.e.,
    normalize to have zero mean) the data without explicitly computing phi(x).
    It is equivalent to centering phi(x) with
    sklearn.preprocessing.StandardScaler(with_std=False).

    Parameters
    ----------
    force_cpu : bool
        If True, computations will be performed on CPU, ignoring the
        current backend. If False, use the current backend.

    Attributes
    ----------
    K_fit_rows_ : array of shape (n_samples,)
        Average of each column of kernel matrix.

    K_fit_all_ : float
        Average of kernel matrix.

    Examples
    --------
    >>> from himalaya.kernel_ridge import KernelCenterer
    >>> from himalaya.kernel_ridge import pairwise_kernels
    >>> X = [[ 1., -2.,  2.],
    ...      [ -2.,  1.,  3.],
    ...      [ 4.,  1., -2.]]
    >>> K = pairwise_kernels(X, metric='linear')
    >>> K
    array([[  9.,   2.,  -2.],
           [  2.,  14., -13.],
           [ -2., -13.,  21.]])
    >>> transformer = KernelCenterer().fit(K)
    >>> transformer
    KernelCenterer()
    >>> transformer.transform(K)
    array([[  5.,   0.,  -5.],
           [  0.,  14., -14.],
           [ -5., -14.,  19.]])
    """

    def __init__(self, force_cpu=False):
        self.force_cpu = force_cpu

    @force_cpu_backend
    def fit(self, K, y=None):
        """Fit KernelCenterer

        Parameters
        ----------
        K : ndarray of shape (n_samples, n_samples)
            Kernel matrix.

        y : None
            Ignored.

        Returns
        -------
        self : object
            Fitted transformer.
        """
        backend = get_backend()
        K = check_array(K, ndim=2)

        if K.shape[0] != K.shape[1]:
            raise ValueError("Kernel matrix must be a square matrix."
                             " Input is a {}x{} matrix.".format(
                                 K.shape[0], K.shape[1]))

        self.K_fit_rows_ = backend.mean_float64(K, axis=0)
        self.K_fit_all_ = backend.mean_float64(self.K_fit_rows_, axis=0)
        return self

    @force_cpu_backend
    def transform(self, K, copy=True):
        """Center kernel matrix.

        Parameters
        ----------
        K : ndarray of shape (n_samples1, n_samples2)
            Kernel matrix.

        copy : bool, default=True
            Set to False to perform inplace computation.

        Returns
        -------
        K_new : ndarray of shape (n_samples1, n_samples2)
        """
        check_is_fitted(self)
        backend = get_backend()
        K = check_array(K, ndim=2, copy=copy)

        K_pred_cols = backend.mean_float64(K, axis=1, keepdims=True)

        K -= self.K_fit_rows_
        K -= K_pred_cols
        K += self.K_fit_all_

        return K

    def _more_tags(self):
        return {'pairwise': True}

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.pairwise = True
        return tags



from . import backend, kernel_ridge, lasso, progress_bar, ridge, scoring, utils, viz

__version__ = "0.4.10"

__all__ = [
    backend,
    kernel_ridge,
    lasso,
    ridge,
    progress_bar,
    scoring,
    utils,
    viz,
]



import numbers

from ..backend import get_backend
from ..progress_bar import ProgressBar
from ..utils import compute_lipschitz_constants
from ..scoring import l2_neg_loss
from ..validation import check_cv

from ._solvers import solve_weighted_kernel_ridge_conjugate_gradient
from ._solvers import solve_weighted_kernel_ridge_gradient_descent
from ._solvers import solve_weighted_kernel_ridge_neumann_series
from ._random_search import solve_multiple_kernel_ridge_random_search


def solve_multiple_kernel_ridge_hyper_gradient(
    Ks,
    Y,
    score_func=l2_neg_loss,
    cv=5,
    fit_intercept=False,
    return_weights=None,
    Xs=None,
    initial_deltas=0,
    max_iter=10,
    tol=1e-2,
    max_iter_inner_dual=1,
    max_iter_inner_hyper=1,
    cg_tol=1e-3,
    n_targets_batch=None,
    hyper_gradient_method="conjugate_gradient",
    kernel_ridge_method="gradient_descent",
    random_state=None,
    progress_bar=True,
    Y_in_cpu=False,
):
    """Solve bilinear kernel ridge regression with cross-validation.

    The hyper-parameters deltas correspond to::

        log(kernel_weights / ridge_regularization)

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Training kernel for each feature space.
    Y : array of shape (n_samples, n_targets)
        Training target data.
    score_func : callable
        Function used to compute the score of predictions.
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    fit_intercept : boolean
        Whether to fit an intercept. If False, Ks should be centered
        (see KernelCenterer), and Y must be zero-mean over samples.
        Only available if return_weights == 'dual'.
    return_weights : None, 'primal', or 'dual'
        Whether to refit on the entire dataset and return the weights.
    Xs : array of shape (n_kernels, n_samples, n_features) or None
        Necessary if return_weights == 'primal'.
    initial_deltas : str, float, array of shape (n_kernels, n_targets)
        Initial log kernel weights for each target.
        If a float, initialize the deltas with this value.
        If a str, initialize the deltas with different strategies:
        - 'ridgecv' : fit a RidgeCV model over the average kernel.
    max_iter : int
        Maximum number of iteration for the outer loop.
    tol : float > 0, or None
        Tolerance for the stopping criterion.
    max_iter_inner_dual : int
        Maximum number of iterations for the dual weights conjugate gradient.
    max_iter_inner_hyper :
        Maximum number of iterations for the deltas gradient descent.
    cg_tol : float, or array of shape (max_iter)
        Tolerance for the conjugate gradients.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    hyper_gradient_method : str, "conjugate_gradient", "neumann", "direct"
        Method to compute the hypergradient.
    kernel_ridge_method : str, "conjugate_gradient" or "gradient_descent"
        Algorithm used for the inner step.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.
    Y_in_cpu : bool
        If True, keep the target values ``Y`` in CPU memory (slower).

    Returns
    -------
    deltas : array of shape (n_kernels, n_targets)
        Best log kernel weights for each target.
    refit_weights : array or None
        Refit regression weights on the entire dataset, using selected best
        hyperparameters. Refit weights will always be on CPU memory.
        If return_weights == 'primal', shape is (n_features, n_targets),
        if return_weights == 'dual', shape is (n_samples, n_targets),
        else, None.
    cv_scores : array of shape (max_iter * max_iter_inner_hyper, n_targets)
        Cross-validation scores per iteration, averaged over splits.
        Cross-validation scores will always be on CPU memory.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()

    n_samples, n_targets = Y.shape
    if n_targets_batch is None:
        n_targets_batch = n_targets

    Ks = backend.asarray(Ks)
    dtype = Ks.dtype
    device = getattr(Ks, "device", None)
    Y = backend.asarray(Y, dtype=dtype, device="cpu" if Y_in_cpu else device)

    if fit_intercept:
        raise NotImplementedError("Coming soon.")

    cv = check_cv(cv, Y)
    n_splits = cv.get_n_splits()
    for train, val in cv.split(Y):
        if len(val) == 0 or len(train) == 0:
            raise ValueError(
                "Empty train or validation set. "
                "Check that `cv` is correctly defined."
            )

    deltas = _init_multiple_kernel_ridge(
        Ks, Y, initial_deltas, cv, n_targets_batch=n_targets_batch, Y_in_cpu=Y_in_cpu
    )

    if return_weights == "primal":
        if Xs is None:
            raise ValueError("Xs is needed to compute the primal weights.")
        n_features = sum(X.shape[1] for X in Xs)
        refit_weights = backend.zeros_like(
            Ks, shape=(n_features, n_targets), device="cpu"
        )

    elif return_weights == "dual":
        refit_weights = backend.zeros_like(
            Ks, shape=(n_samples, n_targets), device="cpu"
        )
    elif return_weights is None:
        refit_weights = None
    else:
        raise ValueError("Unknown parameter return_weights=%r." % (return_weights,))

    lipschitz_constants = None
    name = "hypergradient_" + hyper_gradient_method
    if kernel_ridge_method == "conjugate_gradient":
        inner_function = solve_weighted_kernel_ridge_conjugate_gradient
    elif kernel_ridge_method == "gradient_descent":
        inner_function = solve_weighted_kernel_ridge_gradient_descent

        # precompute the lipschitz constants
        lipschitz_constants = []
        for train, _ in cv.split(Y):
            train = backend.to_gpu(train, device=device)
            Ks_train = Ks[:, train[:, None], train]
            lipschitz_constants.append(
                compute_lipschitz_constants(Ks_train, random_state=random_state)
            )
    else:
        raise ValueError(
            "Unknown parameter kernel_ridge_method=%r." % (kernel_ridge_method,)
        )

    if isinstance(cg_tol, (int, float)):
        cg_tol = backend.full_like(Y, shape=(max_iter,), fill_value=cg_tol)

    alpha = 1.0

    cv_scores = backend.zeros_like(
        Ks, shape=(max_iter * max_iter_inner_hyper, n_targets), device="cpu"
    )

    batch_iterates = range(0, n_targets, n_targets_batch)
    bar = None
    if progress_bar:
        bar = ProgressBar(title=name, max_value=len(batch_iterates) * max_iter)
    for bb, start in enumerate(batch_iterates):
        batch = slice(start, start + n_targets_batch)
        Y_batch = backend.to_gpu(Y[:, batch], device=device)

        previous_solutions = [None] * n_splits
        step_sizes = [None] * n_splits
        dual_weights_cv = [None] * n_splits

        for ii in range(max_iter):
            if progress_bar:
                bar.update(bb * max_iter + ii)

            ##########################
            # updates the dual weights

            # First pass needs more iterations to have something reasonable.
            # We also use conjugate gradient as it converges faster.
            if ii == 0:
                max_iter_inner_dual_ = 50
                cg_tol_ = min(1e-2, cg_tol[ii])
                inner_function_ = solve_weighted_kernel_ridge_conjugate_gradient  # noqa
            else:
                max_iter_inner_dual_ = max_iter_inner_dual
                cg_tol_ = cg_tol[ii]
                inner_function_ = inner_function

            for kk, (train, val) in enumerate(cv.split(Y)):
                train = backend.to_gpu(train, device=device)
                Ks_train = Ks[:, train[:, None], train]
                Y_train = Y_batch[train]

                if kernel_ridge_method == "gradient_descent" and ii != 0:
                    kwargs = dict(lipschitz_Ks=lipschitz_constants[kk])
                else:
                    kwargs = dict()

                dual_weights_cv[kk] = inner_function_(
                    Ks_train,
                    Y_train,
                    deltas[:, batch],
                    initial_dual_weights=dual_weights_cv[kk],
                    alpha=alpha,
                    max_iter=max_iter_inner_dual_,
                    tol=cg_tol_,
                    **kwargs,
                )

            ###################
            # update the deltas
            deltas_old = backend.copy(deltas[:, batch])
            for jj in range(max_iter_inner_hyper):
                gradients = backend.zeros_like(deltas[:, batch])
                scores = backend.zeros_like(
                    Ks, shape=(n_splits, deltas[:, batch].shape[1])
                )
                for kk, (train, val) in enumerate(cv.split(Y)):
                    val = backend.to_gpu(val, device=device)
                    train = backend.to_gpu(train, device=device)

                    Ks_val = Ks[:, val[:, None], train]
                    Ks_train = Ks[:, train[:, None], train]
                    Y_val = Y_batch[val]

                    (
                        gradients_kk,
                        step_sizes[kk],
                        predictions,
                        previous_solutions[kk],
                    ) = _compute_delta_gradient(
                        Ks_val=Ks_val,
                        Y_val=Y_val,
                        deltas=deltas[:, batch],
                        dual_weights=dual_weights_cv[kk],
                        Ks_train=Ks_train,
                        tol=cg_tol[ii],
                        random_state=random_state,
                        hyper_gradient_method=hyper_gradient_method,
                        previous_solution=previous_solutions[kk],
                    )

                    gradients += gradients_kk * Y_val.shape[0] / n_samples

                    scores[kk] = score_func(Y_val, predictions)

                it = ii * max_iter_inner_hyper + jj
                cv_scores[it, batch] = backend.to_cpu(scores.mean(0))

                # update deltas, using the minimum step size over splits
                step_size = backend.min(backend.stack(step_sizes), axis=0)
                deltas[:, batch] -= gradients * step_size[None, :]
                assert not backend.any(backend.isinf(backend.exp(deltas[:, batch])))

            ####################
            # stopping criterion
            if tol is not None:
                if backend.max(backend.abs(deltas_old - deltas[:, batch])) < tol:
                    if it + 1 < cv_scores.shape[0]:
                        cv_scores[it + 1, batch] = cv_scores[it, batch]
                    break

        ##########################################
        # refit dual weights on the entire dataset
        if return_weights in ["primal", "dual"]:
            dual_weights = solve_weighted_kernel_ridge_conjugate_gradient(
                Ks,
                Y_batch,
                deltas[:, batch],
                initial_dual_weights=None,
                alpha=alpha,
                max_iter=100,
                tol=1e-4,
            )
            if return_weights == "primal":
                # multiply by g and not np.sqrt(g), as we then want to use
                # the primal weights on the unscaled features Xs, and not
                # on the scaled features (np.sqrt(g) * Xs)
                X = None
                for tt in range(refit_weights[:, batch].shape[1]):
                    X = backend.concatenate(
                        [
                            t * g
                            for t, g in zip(Xs, backend.exp(deltas[:, batch][:, tt]))
                        ],
                        1,
                    )
                    refit_weights[:, batch][:, tt] = backend.to_cpu(
                        backend.matmul(X.T, dual_weights[:, tt])
                    )
                del X

            elif return_weights == "dual":
                refit_weights[:, batch] = backend.to_cpu(dual_weights)

            del dual_weights

    if progress_bar:
        bar.update(bar.max_value)
        bar.close()

    results = [deltas, refit_weights, cv_scores]
    if fit_intercept:
        pass  # results.append(intercept)
    return results


#: Dictionary with all multiple kernel ridge solvers
MULTIPLE_KERNEL_RIDGE_SOLVERS = {
    "random_search": solve_multiple_kernel_ridge_random_search,
    "hyper_gradient": solve_multiple_kernel_ridge_hyper_gradient,
}


def _init_multiple_kernel_ridge(Ks, Y, initial_deltas, cv, **ridgecv_kwargs):
    """Initialize deltas, i.e. log kernel weights.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples, n_samples)
        Training kernel for each feature space.
    Y : array of shape (n_samples, n_targets)
        Training target data.
    initial_deltas : str, float, array of shape (n_kernels, n_targets)
        Initial log kernel weights for each target.
        If a float, initialize the deltas with this value.
        If a str, initialize the deltas with different strategies:
        - 'ridgecv' : fit a RidgeCV model over the average kernel
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    ridgecv_kwargs : dict
        Other parameters if initial_deltas == 'ridgecv'.

    Returns
    -------
    deltas : array of shape (n_kernels, n_targets)
        Initial deltas.
    """
    backend = get_backend()

    n_kernels = Ks.shape[0]
    n_targets = Y.shape[1]

    if initial_deltas is None:
        initial_deltas = 0

    if isinstance(initial_deltas, str) and (initial_deltas == "ridgecv"):
        alphas = backend.logspace(-10, 20, 30)
        gammas = backend.full_like(Y, shape=n_kernels, fill_value=1.0 / n_kernels)[None]
        deltas, _, _ = solve_multiple_kernel_ridge_random_search(
            Ks,
            Y,
            n_iter=gammas,
            alphas=alphas,
            cv=cv,
            n_alphas_batch=5,
            return_weights=None,
            progress_bar=False,
            **ridgecv_kwargs,
        )

    elif isinstance(initial_deltas, numbers.Number):
        deltas = backend.full_like(
            Ks, shape=(n_kernels, n_targets), fill_value=initial_deltas
        )

    else:
        deltas = backend.copy(backend.asarray_like(initial_deltas, ref=Ks))

    Ks, deltas = backend.check_arrays(Ks, deltas)
    return deltas


def _compute_delta_loss(Ks_val, Y_val, deltas, dual_weights):
    """Compute the validation loss.

    Parameters
    ----------
    Ks_val : array of shape (n_kernels, n_samples_val, n_samples_train)
        Cross-kernels between training set and validation set.
    Y_val : array of shape (n_samples_val, n_targets)
        Target data on the validation set.
    deltas : array of shape (n_kernels, n_targets)
        Log of the kernel weights.
    dual_weights : array of shape (n_samples_train, n_targets)
        Kernel ridge weights.

    Returns
    -------
    loss : array of shape (n_targets)
        L2 validation loss.
    """
    backend = get_backend()
    exp_delta = backend.exp(deltas)
    chi_val = backend.matmul(Ks_val, dual_weights)
    exp_delta_chi_val = exp_delta[:, None, :] * chi_val
    predictions = exp_delta_chi_val.sum(0)
    residuals = predictions - Y_val
    loss = 0.5 * backend.norm(residuals, axis=0) ** 2
    return loss


def _compute_delta_gradient(
    Ks_val,
    Y_val,
    deltas,
    dual_weights,
    Ks_train=None,
    tol=None,
    previous_solution=None,
    hyper_gradient_method="conjugate_gradient",
    random_state=None,
):
    """Compute the gradient over deltas on the validation dataset.

    Parameters
    ----------
    Ks_val : array of shape (n_kernels, n_samples_val, n_samples_train)
        Cross-kernels between training set and validation set.
    Y_val : array of shape (n_samples_val, n_targets)
        Target data on the validation set.
    deltas : array of shape (n_kernels, n_targets)
        Log of the kernel weights.
    dual_weights : array of shape (n_samples_train, n_targets)
        Kernel ridge weights.
    Ks_train : array of shape (n_kernels, n_samples_train, n_samples_train)
        Kernels on the training set.
        Not required if hyper_gradient_method = "direct".
    tol : float
        Tolerance for the conjugate method.
        Required if hyper_gradient_method = "conjugate_gradient".
    previous_solution
        Speed up hyper_gradient_method = "conjugate_gradient" by warm starting
        with the previous solution.
    hyper_gradient_method : str, in {"conjugate_gradient", "neumann", "direct"}
        Method used to compute the hyper gradient.
    random_state : int, or None
        Random generator seed. Use an int for deterministic search.

    Returns
    -------
    gradient : array of shape (n_kernels, n_targets)
        Gradient over deltas.
    step_size : array of shape (n_targets)
        Step size computed based on the direct gradient's Lipschitz constant.
    predictions : array of shape (n_samples_val, n_targets)
        Predictions on the validation set.
    solution : array of shape (n_samples_train, n_targets) or None
        Solution of the inverse Hessian in the indirect gradient.
    """
    backend = get_backend()

    # prepare quantities
    exp_delta = backend.exp(deltas)
    chi_val = backend.matmul(Ks_val, dual_weights)
    exp_delta_chi_val = exp_delta[:, None, :] * chi_val
    predictions = exp_delta_chi_val.sum(0)
    assert predictions.shape == Y_val.shape

    # direct gradient
    residuals = predictions - Y_val
    direct_gradient = (residuals[None] * exp_delta_chi_val).sum(1)
    assert direct_gradient.shape == deltas.shape

    # estimate a step size
    XTXs = _compute_deltas_hessian(exp_delta_chi_val, residuals)
    # (these lipschitz constants only correspond to the direct gradient)
    lipschitz_1 = compute_lipschitz_constants(XTXs, "X", random_state=random_state)
    step_size = 1.0 / (lipschitz_1 + 1e-15)

    if hyper_gradient_method == "direct":
        gradient = direct_gradient
        solution = None
    else:
        # compute the indirect gradient by inverting the Hessian

        # compute nabla_g_1
        tmp = backend.matmul(backend.transpose(Ks_val, (2, 0, 1)), residuals)
        tmp = backend.transpose(tmp, (2, 0, 1))
        nabla_g_1 = backend.matmul(
            tmp, backend.transpose(exp_delta, (1, 0))[:, :, None]
        )
        nabla_g_1 = backend.transpose(nabla_g_1[:, :, 0], (1, 0))
        assert nabla_g_1.shape == dual_weights.shape

        # solve linear system (sum_i gamma[i]*K[i] + 1) @ X = nabla_g_1
        alpha = 1
        assert Ks_train is not None
        if hyper_gradient_method == "conjugate_gradient":
            assert tol is not None
            solution = solve_weighted_kernel_ridge_conjugate_gradient(
                Ks=Ks_train,
                Y=nabla_g_1,
                deltas=deltas,
                initial_dual_weights=previous_solution,
                max_iter=100,
                tol=tol,
                alpha=alpha,
            )
        elif hyper_gradient_method == "neumann":
            solution = solve_weighted_kernel_ridge_neumann_series(
                Ks=Ks_train,
                Y=nabla_g_1,
                deltas=deltas,
                max_iter=5,
                factor=0.00001,
                alpha=alpha,
            )
        else:
            raise ValueError(
                "Unknown parameter hyper_gradient_method=%r." % (hyper_gradient_method,)
            )

        # finish the indirect gradient
        chi_train = backend.matmul(Ks_train, dual_weights)
        exp_delta_chi_train = exp_delta[:, None, :] * chi_train
        indirect_gradient = (exp_delta_chi_train * solution[None, :, :]).sum(1)
        assert indirect_gradient.shape == deltas.shape

        gradient = direct_gradient - indirect_gradient
        assert not backend.any(backend.isinf(gradient))

    return gradient, step_size, predictions, solution


def _compute_deltas_hessian(exp_delta_chi, residuals):
    """Compute the hessian of the direct gradient.

    The direct gradient corresponds to a linear problem:
        argmin_delta ||exp(delta) @ chi - Y||^2
    where chi = Ks @ dual_weights.

    The Hessian is not just `chi.T @ chi` because of the exponential
    parametrization of deltas.

    Parameters
    ----------
    exp_delta_chi : array of shape (n_kernels, n_samples, n_targets)
        Precomputation of exp(delta) * (Ks @ dual_weights).
    residuals : array of shape (n_samples, n_targets)
        Difference between prediction and target on the validation split.

    Returns
    -------
    XTXs : array of shape (n_targets, n_kernels, n_kernels)
        Hessian of the direct gradient.
    """
    backend = get_backend()

    XTXs = backend.matmul(
        backend.transpose(exp_delta_chi, (2, 0, 1)),
        backend.transpose(exp_delta_chi, (2, 1, 0)),
    )
    XTbs = backend.matmul(
        backend.transpose(exp_delta_chi, (2, 0, 1)),
        backend.transpose(residuals, (1, 0))[:, :, None],
    )[:, :, 0]
    diagonal_view = backend.diagonal_view(XTXs, axis1=1, axis2=2)
    diagonal_view += XTbs
    return XTXs



"""The "numpy" CPU backend, based on NumPy.

To use this backend, call ``himalaya.backend.set_backend("numpy")``.
"""
import numpy as np
try:
    import scipy.linalg as linalg
    use_scipy = True
except ImportError:
    import numpy.linalg as linalg
    use_scipy = False

###############################################################################


def apply_argmax(array, argmax, axis):
    """Apply precomputed argmax indices in multi dimension arrays

    array[np.argmax(array)] works fine in dimension 1, but not in higher ones.
    This function extends it to higher dimensions.

    Examples
    --------
    >>> import numpy as np
    >>> array = np.random.randn(10, 4, 8)
    >>> argmax = np.argmax(array, axis=1)
    >>> max_ = apply_argmax(array, argmax, axis=1)
    >>> assert np.all(max_ == np.max(array, axis=1))
    """
    argmax = np.expand_dims(argmax, axis=axis)
    max_ = np.take_along_axis(array, argmax, axis=axis)
    return np.take(max_, 0, axis=axis)


def std_float64(array, axis=None, demean=True, keepdims=False):
    """Compute the standard deviation of X with double precision,
    and cast back the result to original dtype.
    """
    return array.std(axis, dtype=np.float64,
                     keepdims=keepdims).astype(array.dtype, copy=False)


def mean_float64(array, axis=None, keepdims=False):
    """Compute the mean of X with double precision,
    and cast back the result to original dtype.
    """
    return array.mean(axis, dtype=np.float64,
                      keepdims=keepdims).astype(array.dtype, copy=False)


###############################################################################

name = "numpy"
argmax = np.argmax
max = np.max
min = np.min
abs = np.abs
randn = np.random.randn
rand = np.random.rand
matmul = np.matmul
transpose = np.transpose
stack = np.stack
concatenate = np.concatenate
sum = np.sum
sqrt = np.sqrt
any = np.any
all = np.all
nan = np.nan
inf = np.inf
isnan = np.isnan
isinf = np.isinf
logspace = np.logspace
copy = np.copy
bool = np.bool_
float32 = np.float32
float64 = np.float64
int32 = np.int32
eigh = linalg.eigh
norm = linalg.norm
log = np.log
exp = np.exp
arange = np.arange
flatnonzero = np.flatnonzero
isin = np.isin
searchsorted = np.searchsorted
unique = np.unique
einsum = np.einsum
tanh = np.tanh
power = np.power
prod = np.prod
zeros = np.zeros
clip = np.clip
sign = np.sign
sort = np.sort
flip = np.flip
atleast_1d = np.atleast_1d
finfo = np.finfo
eye = np.eye


def diagonal_view(array, axis1=0, axis2=1):
    """Return a view of the array diagonal"""
    assert array.ndim >= 2
    axis1, axis2 = min([axis1, axis2]), max([axis1, axis2])
    shape = list(array.shape)
    new = min([shape[axis1], shape[axis2]])
    shape.pop(axis1)
    shape.pop(axis2 - 1)
    shape.append(new)
    strides = list(array.strides)
    new = strides[axis1] + strides[axis2]
    strides.pop(axis1)
    strides.pop(axis2 - 1)
    strides.append(new)
    diag = np.lib.stride_tricks.as_strided(array, shape=shape, strides=strides)
    return diag


def to_numpy(array):
    return array


def zeros_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in zeros_like."""
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    return np.zeros(shape, dtype=dtype)


def ones_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in ones_like."""
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    return np.ones(shape, dtype=dtype)


def full_like(array, fill_value, shape=None, dtype=None, device=None):
    """Add a shape parameter in full_like."""
    if shape is None:
        shape = array.shape
    if dtype is None:
        dtype = array.dtype
    return np.full(shape, fill_value, dtype=dtype)


def to_cpu(array):
    return array


def to_gpu(array, device=None):
    return array


def is_in_gpu(array):
    return False


def asarray_like(x, ref):
    return np.asarray(x, dtype=ref.dtype)


def check_arrays(*all_inputs):
    """Change all inputs into arrays (or list of arrays) using the same
    precision as the first one. Some arrays can be None.
    """
    all_arrays = []
    all_arrays.append(asarray(all_inputs[0]))
    dtype = all_arrays[0].dtype
    for tensor in all_inputs[1:]:
        if tensor is None:
            pass
        elif isinstance(tensor, list):
            tensor = [asarray(tt, dtype=dtype) for tt in tensor]
        else:
            tensor = asarray(tensor, dtype=dtype)
        all_arrays.append(tensor)
    return all_arrays


def asarray(a, dtype=None, order=None, device=None):
    # works from numpy, lists, torch, and others
    try:
        return np.asarray(a, dtype=dtype, order=order)
    except Exception:
        pass
    # works from cupy
    try:
        import cupy
        return np.asarray(cupy.asnumpy(a), dtype=dtype, order=order)
    except Exception:
        pass
    # works from torch_cuda
    try:
        return np.asarray(a.cpu(), dtype=dtype, order=order)
    except Exception:
        pass

    return np.asarray(a, dtype=dtype, order=order)


def svd(X, full_matrices=True):
    if X.ndim == 2 or not use_scipy:
        return linalg.svd(X, full_matrices=full_matrices)

    elif X.ndim == 3:
        UsV_list = [linalg.svd(Xi, full_matrices=full_matrices) for Xi in X]
        return map(np.stack, zip(*UsV_list))
    else:
        raise NotImplementedError()



from ._solvers import solve_weighted_kernel_ridge_gradient_descent
from ._solvers import solve_weighted_kernel_ridge_conjugate_gradient
from ._solvers import solve_weighted_kernel_ridge_neumann_series
from ._solvers import solve_kernel_ridge_eigenvalues
from ._solvers import solve_kernel_ridge_gradient_descent
from ._solvers import solve_kernel_ridge_conjugate_gradient
from ._solvers import KERNEL_RIDGE_SOLVERS
from ._solvers import WEIGHTED_KERNEL_RIDGE_SOLVERS
from ._hyper_gradient import solve_multiple_kernel_ridge_hyper_gradient
from ._hyper_gradient import MULTIPLE_KERNEL_RIDGE_SOLVERS
from ._random_search import solve_multiple_kernel_ridge_random_search
from ._random_search import generate_dirichlet_samples
from ._random_search import solve_kernel_ridge_cv_eigenvalues
from ._random_search import solve_kernel_ridge_cv_svd
from ._random_search import KERNEL_RIDGE_CV_SOLVERS
from ._predictions import predict_weighted_kernel_ridge
from ._predictions import predict_and_score_weighted_kernel_ridge
from ._predictions import primal_weights_kernel_ridge
from ._predictions import primal_weights_weighted_kernel_ridge
from ._sklearn_api import KernelRidge
from ._sklearn_api import KernelRidgeCV
from ._sklearn_api import MultipleKernelRidgeCV
from ._sklearn_api import WeightedKernelRidge
from ._kernels import PAIRWISE_KERNEL_FUNCTIONS
from ._kernels import linear_kernel
from ._kernels import polynomial_kernel
from ._kernels import rbf_kernel
from ._kernels import sigmoid_kernel
from ._kernels import cosine_similarity_kernel
from ._kernels import KernelCenterer
from ._kernelizer import Kernelizer
from ._kernelizer import ColumnKernelizer
from ._kernelizer import make_column_kernelizer

__all__ = [
    # kernel ridge solvers
    "solve_weighted_kernel_ridge_gradient_descent",
    "solve_weighted_kernel_ridge_conjugate_gradient",
    "solve_weighted_kernel_ridge_neumann_series",
    "solve_kernel_ridge_cv_eigenvalues",
    "solve_kernel_ridge_cv_svd",
    "solve_kernel_ridge_eigenvalues",
    "solve_kernel_ridge_gradient_descent",
    "solve_kernel_ridge_conjugate_gradient",
    "KERNEL_RIDGE_SOLVERS",
    "KERNEL_RIDGE_CV_SOLVERS",
    "WEIGHTED_KERNEL_RIDGE_SOLVERS",
    # multiple kernel ridge solvers
    "MULTIPLE_KERNEL_RIDGE_SOLVERS",
    "solve_multiple_kernel_ridge_hyper_gradient",
    "solve_multiple_kernel_ridge_random_search",
    # helpers
    "generate_dirichlet_samples",
    "predict_weighted_kernel_ridge",
    "predict_and_score_weighted_kernel_ridge",
    "primal_weights_kernel_ridge",
    "primal_weights_weighted_kernel_ridge",
    # scikit-learn API
    "KernelRidge",
    "KernelRidgeCV",
    "MultipleKernelRidgeCV",
    "WeightedKernelRidge",
    # kernels
    "PAIRWISE_KERNEL_FUNCTIONS",
    "linear_kernel",
    "polynomial_kernel",
    "rbf_kernel",
    "sigmoid_kernel",
    "cosine_similarity_kernel",
    "KernelCenterer",
    # kernelizers
    "Kernelizer",
    "ColumnKernelizer",
    "make_column_kernelizer",
]



import numbers
import warnings

from ..backend import get_backend
from ..utils import _batch_or_skip


def solve_ridge_svd(X, Y, alpha=1., method="svd", fit_intercept=False,
                    negative_eigenvalues="zeros", n_targets_batch=None,
                    warn=True):
    """Solve ridge regression using SVD decomposition.

    Solve the ridge regression::

        b* = argmin_B ||X @ b - Y||^2 + alpha ||b||^2

    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        Input features.
    Y : array of shape (n_samples, n_targets)
        Target data.
    alpha : float, or array of shape (n_targets, )
        Regularization parameter.
    method : str in {"svd"}
        Method used to diagonalize the input feature matrix.
    fit_intercept : boolean
        Whether to fit an intercept.
        If False, X and Y must be zero-mean over samples.
    negative_eigenvalues : str in {"nan", "error", "zeros"}
        If the decomposition leads to negative eigenvalues (wrongly emerging
        from float32 errors):
        - "error" raises an error.
        - "zeros" replaces them with zeros.
        - "nan" returns nans if the regularization does not compensate
        twice the smallest negative value, else it ignores the problem.
    n_targets_batch : int or None
        Size of the batch for over targets during cross-validation.
        Used for memory reasons. If None, uses all n_targets at once.
    warn : bool
        If True, warn if the number of samples is smaller than the number of
        features.

    Returns
    -------
    weights : array of shape (n_features, n_targets)
        Ridge coefficients.
    intercept : array of shape (n_targets,)
        Intercept. Only returned when fit_intercept is True.
    """
    backend = get_backend()
    if isinstance(alpha, numbers.Number) or alpha.ndim == 0:
        alpha = backend.ones_like(Y, shape=(1, )) * alpha

    X, Y, alpha = backend.check_arrays(X, Y, alpha)

    n_samples, n_features = X.shape
    if n_samples < n_features and warn:
        warnings.warn(
            "Solving ridge is slower than solving kernel ridge when n_samples "
            f"< n_features (here {n_samples} < {n_features}). "
            "Using a linear kernel in himalaya.kernel_ridge.KernelRidge or "
            "himalaya.kernel_ridge.solve_kernel_ridge_eigenvalues would be "
            "faster. Use warn=False to silence this warning.", UserWarning)
    if X.shape[0] != Y.shape[0]:
        raise ValueError("X and Y must have the same number of samples.")

    X_offset, Y_offset = None, None
    if fit_intercept:
        X_offset = X.mean(0)
        Y_offset = Y.mean(0)
        X = X - X_offset
        Y = Y - Y_offset

    if method == "svd":
        # SVD: X = U @ np.diag(eigenvalues) @ Vt
        U, eigenvalues, Vt = backend.svd(X, full_matrices=False)
    else:
        raise ValueError("Unknown method=%r." % (method, ))

    inverse = eigenvalues[:, None] / (alpha[None] + eigenvalues[:, None] ** 2)

    # negative eigenvalues can emerge from incorrect kernels, or from float32
    if eigenvalues[0] < 0:
        if negative_eigenvalues == "nan":
            if alpha < -eigenvalues[0] * 2:
                return backend.ones_like(Y) * backend.asarray(
                    backend.nan, dtype=Y.dtype)
            else:
                pass

        elif negative_eigenvalues == "zeros":
            eigenvalues[eigenvalues < 0] = 0

        elif negative_eigenvalues == "error":
            raise RuntimeError(
                "Negative eigenvalues. Make sure the kernel is positive "
                "semi-definite, increase the regularization alpha, or use"
                "another solver.")
        else:
            raise ValueError("Unknown negative_eigenvalues=%r." %
                             (negative_eigenvalues, ))

    n_samples, n_features = X.shape
    n_samples, n_targets = Y.shape
    weights = backend.zeros_like(X, shape=(n_features, n_targets),
                                 device="cpu")
    if n_targets_batch is None:
        n_targets_batch = n_targets

    for start in range(0, n_targets, n_targets_batch):
        batch = slice(start, start + n_targets_batch)

        iUT = _batch_or_skip(inverse, batch, 1)[:, None, :] * U.T[:, :, None]
        iUT = backend.transpose(iUT, (2, 0, 1))
        # iUT.shape = (1 or n_targets_batch, n_samples, n_samples)

        if Y.shape[0] < Y.shape[1]:
            weights_batch = ((Vt.T @ iUT) @ Y.T[batch, :, None])[:, :, 0].T
        else:
            weights_batch = Vt.T @ (iUT @ Y.T[batch, :, None])[:, :, 0].T
        weights[:, batch] = backend.to_cpu(weights_batch)

    if fit_intercept:
        intercept = backend.to_cpu(
            Y_offset) - backend.to_cpu(X_offset) @ weights
        return weights, intercept
    else:
        return weights


#: Dictionary with all ridge solvers
RIDGE_SOLVERS = {"svd": solve_ridge_svd}



import numbers
import warnings

import numpy as np
try:
    from numpy.core.numeric import ComplexWarning
except ImportError:
    from numpy.exceptions import ComplexWarning

try:
    from scipy.sparse import issparse
except ImportError:

    def issparse(X):
        return False


from sklearn import get_config
from sklearn.model_selection import check_cv as check_cv_sklearn

from .backend import get_backend
from .backend._utils import _dtype_to_str


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


def check_array(array, accept_sparse=False, dtype=["float32", "float64"],
                copy=False, force_all_finite=True, ndim=2,
                ensure_min_samples=1, ensure_min_features=1, device="default"):
    """Input validation on an array, list, sparse matrix or similar.

    By default, the input is checked to be a non-empty 2D array containing
    only finite values. If the dtype of the array is object, attempt
    converting to float, raising on failure.

    Parameters
    ----------
    array : object
        Input object to check / convert.

    accept_sparse : string, boolean or list/tuple of strings (default=False)
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc. If the input is sparse but not in the allowed format,
        it will be converted to the first listed format. True allows the input
        to be any format. False means that a sparse matrix input will
        raise an error.

    dtype : str, list of str or None
        Data type of result. If None, the dtype of the input is preserved.
        If dtype is a list of str, conversion on the first type is only
        performed if the dtype of the input is not in the list.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in array. The
        possibilities are:

        - True: Force all values of array to be finite.
        - False: accept both np.inf and np.nan in array.
        - 'allow-nan': accept only np.nan values in array. Values cannot
          be infinite.

    ndim : int, list of int, or None (default=2)
        If not None, list the accepted number of dimensions of the array.

    ensure_min_samples : int (default=1)
        Make sure that the array has a minimum number of samples in its first
        axis (rows for a 2D array). Setting to 0 disables this check.

    ensure_min_features : int (default=1)
        Make sure that the 2D array has some minimum number of features
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when the input data has effectively 2
        dimensions or is originally 1D and ``ensure_2d`` is True. Setting to 0
        disables this check.

    device : misc (default="default")
        Parameter passed to ``backend.asarray``. If "default", does not pass
        any parameter, to get the function's default. If "cpu", force the
        array to be in CPU memory.

    Returns
    -------
    array_converted : object
        The converted and validated array.
    """
    backend = get_backend()

    ########################
    # decide to change dtype

    dtype_input = _get_string_dtype(array)

    if dtype is None:
        dtype = None
    elif isinstance(dtype, str):
        if dtype not in ("float32", "float64"):
            raise ValueError("Unknown dtype=%r" % dtype)
    elif isinstance(dtype, (list, tuple)):
        if dtype_input in dtype:
            dtype = dtype_input
        else:
            dtype = dtype[0]
    else:
        raise ValueError("Unknown dtype=%r" % dtype)

    if not isinstance(device, str) or device != "default":
        kwargs = dict(device=device)
    else:
        kwargs = dict()

    if array is None:
        raise ValueError(
            "Expected array-like (array or non-string sequence), got None")

    #############
    # sparse case
    if issparse(array):
        changed_format = False

        if isinstance(accept_sparse, str):
            accept_sparse = [accept_sparse]
        if accept_sparse is False:
            raise TypeError('A sparse matrix was passed, but dense '
                            'data is required. Use X.toarray() to '
                            'convert to a dense numpy array.')
        elif isinstance(accept_sparse, (list, tuple)):
            if len(accept_sparse) == 0:
                raise ValueError("When providing 'accept_sparse' "
                                 "as a tuple or list, it must contain at "
                                 "least one string value.")
            # ensure correct sparse format
            if array.format not in accept_sparse:
                array = array.asformat(accept_sparse[0])
                changed_format = True
        elif accept_sparse is not True:
            raise ValueError("Parameter 'accept_sparse' should be a string, "
                             "boolean or list of strings. You provided "
                             "'accept_sparse={}'.".format(accept_sparse))

        if dtype != dtype_input:
            array = array.astype(dtype)
        elif copy and not changed_format:
            array = array.copy()

        _assert_all_finite(array.data, force_all_finite, numpy=True)

    ############
    # dense case
    else:

        if hasattr(array, 'dtype') and "complex" in str(array.dtype):
            raise ValueError("Complex data not supported.")
        # convert ComplexWarning into error
        with warnings.catch_warnings():
            try:
                warnings.simplefilter('error', ComplexWarning)
                ################
                array = backend.asarray(array, dtype=dtype, **kwargs)
                ################
            except ComplexWarning:
                raise ValueError("Complex data not supported.")

        if copy:
            array = backend.copy(array)

        if ndim is not None and array.ndim not in np.atleast_1d(ndim):
            raise ValueError("Found array with ndim %d, expected %s. "
                             "Reshape your data or change the input." %
                             (array.ndim, ndim))

        # copy misaligned arrays, as it can lead to segmentation faults
        if hasattr(array, "data_ptr"):
            if array.data_ptr() % 8 != 0:
                array = backend.copy(array)
        if hasattr(array, "__array_interface__"):
            if array.__array_interface__['data'][0] % 8 != 0:
                array = backend.copy(array)

        numpy = backend.name == 'cupy' and device == "cpu"
        _assert_all_finite(array, force_all_finite, numpy=numpy)

    #####################
    # check minimum sizes
    if ensure_min_samples > 0:
        n_samples = array.shape[0]
        if n_samples < ensure_min_samples:
            raise ValueError(
                "Found array with %d sample(s) (shape=%s) while a"
                " minimum of %d is required." %
                (n_samples, tuple(array.shape), ensure_min_samples))

    if ensure_min_features > 0 and array.ndim == 2:
        n_features = array.shape[1]
        if n_features < ensure_min_features:
            raise ValueError(
                "Found array with %d feature(s) (shape=%s) while"
                " a minimum of %d is required." %
                (n_features, tuple(array.shape), ensure_min_features))

    return array


def _assert_all_finite(X, force_all_finite, numpy=False, batch_size=2 ** 24):
    """Check infinity and NaNs in X.

    Parameters
    ----------
    X : array
        Input data.
    force_all_finite : {True, False, 'allow-nan'}
        If True, raise an error for infinity and NaNs.
        If False, does not raise any error.
        If 'allow_nan', raise an error for infinity but not for NaNs.
    numpy : bool
        If True, use numpy for the check, else use the current backend.
    batch_size : int
        Batch size to avoid a full memory copy of X.

    Return
    ------
    None
    """
    if numpy:
        backend = np
    else:
        backend = get_backend()

    if get_config()['assume_finite']:
        return

    if force_all_finite not in (True, False, 'allow-nan'):
        raise ValueError('force_all_finite should be a bool or "allow-nan"'
                         '. Got {!r} instead'.format(force_all_finite))
    if not force_all_finite:
        return

    X_dtype = _get_string_dtype(X)
    msg = ("Input contains NaN, infinity or a value too large for "
           "dtype(%r)." % X_dtype)

    for start in range(0, np.prod(X.shape), batch_size):
        batch = slice(start, start + batch_size)
        X_batch = X.flatten()[batch]

        if backend.any(backend.isinf(X_batch)):
            raise ValueError(msg)
        if force_all_finite != 'allow_nan' and backend.any(
                backend.isnan(X_batch)):
            raise ValueError(msg)


def _get_string_dtype(array):
    """Get array's dtype, as a string.
    Returns None, if array has no attribute dtype.
    """
    dtype = getattr(array, "dtype", None)
    if dtype is None:
        return None

    return _dtype_to_str(dtype)


def validate_data(estimator, X='no_validation', y='no_validation', reset=True, feature_axis=-1, **check_params):
    """Input validation for estimators that wraps check_array with n_features_in_ management.

    This function provides sklearn 1.6+ compatible input validation by wrapping
    the existing check_array function and managing the n_features_in_ attribute
    for sklearn compatibility.

    Parameters
    ----------
    estimator : object
        The estimator instance for which to validate data.
    X : array-like, default='no_validation'
        Input data to validate. If 'no_validation', skip X validation.
    y : array-like, default='no_validation'
        Target data to validate. If 'no_validation', skip y validation.
    reset : bool, default=True
        Whether to reset the n_features_in_ attribute. Set to True during fit(),
        False during predict().
    feature_axis : int, default=-1
        Axis along which to count features. Default -1 (last axis) handles
        both 2D arrays (axis 1) and 3D precomputed kernels (axis 2) correctly.
        For regular 2D input: shape (n_samples, n_features) -> axis -1 = axis 1
        For 3D precomputed kernels: shape (n_kernels, n_samples, n_samples) -> axis -1 = axis 2
    **check_params : dict
        Additional parameters to pass to check_array for X validation.

    Returns
    -------
    X_validated : array-like or 'no_validation'
        Validated input data.
    y_validated : array-like or 'no_validation'
        Validated target data.
    """
    # Handle X validation
    validate_X = not (isinstance(X, str) and X == 'no_validation')
    if validate_X:
        X_validated = check_array(X, **check_params)

        # Manage n_features_in_ attribute
        if reset:
            # During fit: store number of features using specified axis
            estimator.n_features_in_ = X_validated.shape[feature_axis]
        else:
            # During predict: check consistency with stored value
            if hasattr(estimator, 'n_features_in_'):
                current_n_features = X_validated.shape[feature_axis]
                if current_n_features != estimator.n_features_in_:
                    raise ValueError(
                        f"X has {current_n_features} features, but {estimator.__class__.__name__} "
                        f"is expecting {estimator.n_features_in_} features as input."
                    )
    else:
        X_validated = X

    # Handle y validation
    validate_y = not (isinstance(y, str) and y == 'no_validation')
    if validate_y:
        # For y, we don't pass all check_params, just extract relevant ones
        y_check_params = {k: v for k, v in check_params.items()
                          if k in ['dtype', 'accept_sparse', 'copy', 'force_all_finite', 'device']}
        # y typically has different ndim requirements
        if 'ndim' not in y_check_params:
            y_check_params['ndim'] = [1, 2]
        y_validated = check_array(y, **y_check_params)
    else:
        y_validated = y

    # Return appropriate format based on what was validated
    if validate_X and validate_y:
        return X_validated, y_validated
    elif validate_X:
        return X_validated
    elif validate_y:
        return y_validated
    else:
        # Return 'no_validation' if only that was passed (matches sklearn behavior)
        if X == 'no_validation' and y == 'no_validation':
            return 'no_validation'
        return None


def check_cv(cv, y):
    """Check the cross-validation generator object.

    In particular, check if indices in cv exceed number of samples, because
    some backends do not check for out-of-bound errors.

    Parameters
    ----------
    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.
    y : array of shape (n_samples,) or (n_samples, n_targets)
        Target values.

    Returns
    -------
    cv : scikit-learn splitter
        Cross-validation splitter.
    """
    # first call check_cv from sklearn
    cv = check_cv_sklearn(cv)

    for train, test in cv.split(y):
        if max(train) >= y.shape[0] or max(test) >= y.shape[0]:
            # some backends do not check out of bound indices
            raise ValueError("Indices in cv exceed number of samples.")
    return cv



from packaging import version

import sklearn
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import (
    ColumnTransformer,
    make_column_selector,  # noqa
)
from sklearn.pipeline import _name_estimators, make_pipeline
from sklearn.utils.validation import check_is_fitted

from ..backend import force_cpu_backend, get_backend
from ..validation import _get_string_dtype, check_array, validate_data
from ._kernels import PAIRWISE_KERNEL_FUNCTIONS, pairwise_kernels


class Kernelizer(TransformerMixin, BaseEstimator):
    """Transform tabular data into a kernel.

    Parameters
    ----------
    kernel : str or callable, default="linear"
        Kernel mapping. Available kernels are: 'linear',
        'polynomial, 'poly', 'rbf', 'sigmoid', 'cosine', or 'precomputed'.
        Set to 'precomputed' in order to pass a precomputed kernel matrix to
        the estimator methods instead of samples.
        A callable should accept two arguments and the keyword arguments passed
        to this object as kernel_params, and should return a floating point
        number.

    kernel_params : dict or None
        Additional parameters for the kernel function.
        See more details in the docstring of the function:
        Kernelizer.ALL_KERNELS[kernel]

    force_cpu : bool
        If True, computations will be performed on CPU, ignoring the
        current backend. If False, use the current backend.

    Attributes
    ----------
    X_fit_ : array of shape (n_samples, n_features)
        Training data. If kernel == "precomputed" this is None.

    n_features_in_ : int
        Number of features (or number of samples if kernel == "precomputed")
        used during the fit.

    dtype_ : str
        Dtype of input data.

    Examples
    --------
    >>> from himalaya.kernel_ridge import Kernelizer
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 3
    >>> X = np.random.randn(n_samples, n_features)
    >>> model = Kernelizer()
    >>> model.fit_transform(X).shape
    (10, 10)
    """

    ALL_KERNELS = PAIRWISE_KERNEL_FUNCTIONS
    kernelizer = True

    def __init__(self, kernel="linear", kernel_params=None, force_cpu=False):
        self.kernel = kernel
        self.kernel_params = kernel_params
        self.force_cpu = force_cpu

    @force_cpu_backend
    def fit_transform(self, X, y=None):
        """Compute the kernel on the training set.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)
            Training data. If kernel == "precomputed" this is instead
            a precomputed kernel array of shape (n_samples, n_samples).

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        K : array of shape (n_samples, n_samples)
            Kernel of the input data.
        """
        accept_sparse = False if self.kernel == "precomputed" else ("csr",
                                                                    "csc")
        X = validate_data(self, X, reset=True, accept_sparse=accept_sparse, ndim=2)

        self.X_fit_ = _to_cpu(X) if self.kernel != "precomputed" else None
        self.dtype_ = _get_string_dtype(X)

        K = self._get_kernel(X)
        return K

    @force_cpu_backend
    def fit(self, X, y=None):
        """Compute the kernel on the training set.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)
            Training data. If kernel == "precomputed" this is instead
            a precomputed kernel array of shape (n_samples, n_samples).

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        self.fit_transform(X, y)
        return self

    @force_cpu_backend
    def transform(self, X):
        """Compute the kernel on any data set.

        Parameters
        ----------
        X : array of shape (n_samples_transform, n_features)
            Training data. If kernel == "precomputed" this is instead
            a precomputed kernel array of shape
            (n_samples_transform, n_samples_fit).

        Returns
        -------
        K : array of shape (n_samples_transform, n_samples_fit)
            Kernel of the input data.
        """
        check_is_fitted(self)
        accept_sparse = False if self.kernel == "precomputed" else ("csr",
                                                                    "csc")
        X = validate_data(self, X, reset=False, dtype=self.dtype_, accept_sparse=accept_sparse,
                        ndim=2)
        K = self._get_kernel(X, self.X_fit_)
        return K

    def _get_kernel(self, X, Y=None):
        """Helper function to get the kernel."""
        backend = get_backend()
        kernel_params = self.kernel_params or {}
        kernel = pairwise_kernels(X, Y, metric=self.kernel, **kernel_params)
        return backend.asarray(kernel)

    def get_X_fit(self):
        """Helper to get the input data X seen during the fit.

        Returns
        -------
        X : array of shape (n_samples, n_features)
            Input array for the kernelizer.
        """
        check_is_fitted(self)
        return self.X_fit_

    @property
    def _pairwise(self):
        return self.kernel == "precomputed"


class ColumnKernelizer(ColumnTransformer):
    """Applies transformers to columns of an array, ending with kernelizers.

    This estimator allows different columns or column subsets of the input
    to be transformed separately. Each transformer pipeline either ends with a
    kernelizer, or a linear kernelizer is added at the end. The different
    kernels generated are then stacked together to be used e.g. in a
    MultipleKernelRidgeCV(kernels="precomputed"). This is useful to perform
    separate transformations and kernels on different feature spaces.

    Warning : This class does not perfectly abide by scikit-learn's API.
    Indeed, it returns stacked kernels of shape (n_kernels, n_samples,
    n_samples), while scikit-learn's API only allows arrays of shape
    (n_samples, n_samples) or (n_samples, n_features). This class is intended
    to be used in a scikit-learn pipeline *just* before a
    MultipleKernelRidgeCV(kernels="precomputed").

    Parameters
    ----------
    transformers : list of tuples
        List of (name, transformer, columns) tuples specifying the
        transformer objects to be applied to subsets of the data.

        name : str
            Like in Pipeline and FeatureUnion, this allows the transformer and
            its parameters to be set using ``set_params`` and searched in grid
            search.
        transformer : {'drop', 'passthrough'} or estimator
            Estimator must support ``fit`` and ``transform``.
            Special-cased strings 'drop' and 'passthrough' are accepted as
            well, to indicate to drop the columns or to pass them through
            untransformed, respectively. If the transformer does not return a
            kernel (as informed by the attribute kernelizer=True), a linear
            kernelizer is applied after the transformer.
        columns :  str, array-like of str, int, array-like of int, \
                array-like of bool, slice or callable
            Indexes the data on its second axis. Integers are interpreted as
            positional columns, while strings can reference DataFrame columns
            by name.  A scalar string or int should be used where
            ``transformer`` expects X to be a 1d array-like (vector),
            otherwise a 2d array will be passed to the transformer.
            A callable is passed the input data `X` and can return any of the
            above. To select multiple columns by name or dtype, you can use
            :obj:`make_column_selector`.

    remainder : {'drop', 'passthrough'} or estimator, default='drop'
        By default, only the specified columns in `transformers` are
        transformed and combined in the output, and the non-specified
        columns are dropped. (default of ``'drop'``).
        By specifying ``remainder='passthrough'``, all remaining columns that
        were not specified in `transformers` will be automatically passed
        through. This subset of columns is concatenated with the output of
        the transformers.
        By setting ``remainder`` to be an estimator, the remaining
        non-specified columns will use the ``remainder`` estimator. The
        estimator must support ``fit`` and ``transform``.
        Note that using this feature requires that the DataFrame columns
        input at ``fit`` and ``transform`` have identical order.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        n_jobs does not work with GPU backends.

    transformer_weights : dict, default=None
        Multiplicative weights for features per transformer. The output of the
        transformer is multiplied by these weights. Keys are transformer names,
        values the weights.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.

    force_cpu : bool
        If True, computations will be performed on CPU, ignoring the
        current backend. If False, use the current backend.

    force_int_remainder_cols : bool, default=True
        Force the columns of the last entry of transformers_, which corresponds 
        to the “remainder” transformer, to always be stored as indices (int) 
        rather than column names (str). 

    Attributes
    ----------
    transformers_ : list
        The collection of fitted transformers as tuples of
        (name, fitted_transformer, column). `fitted_transformer` can be an
        estimator, 'drop', or 'passthrough'. In case there were no columns
        selected, this will be the unfitted transformer.
        If there are remaining columns, the final element is a tuple of the
        form:
        ('remainder', transformer, remaining_columns) corresponding to the
        ``remainder`` parameter. If there are remaining columns, then
        ``len(transformers_)==len(transformers)+1``, otherwise
        ``len(transformers_)==len(transformers)``.

    named_transformers_ : :class:`~sklearn.utils.Bunch`
        Read-only attribute to access any transformer by given name.
        Keys are transformer names and values are the fitted transformer
        objects.

    n_features_in_ : int
        Number of features (or number of samples if kernel == "precomputed")
        used during the fit.

    sparse_output_ : False

    Notes
    -----
    The order of the columns in the transformed feature matrix follows the
    order of how the columns are specified in the `transformers` list.
    Columns of the original feature matrix that are not specified are
    dropped from the resulting transformed feature matrix, unless specified
    in the `passthrough` keyword. Those columns specified with `passthrough`
    are added at the right to the output of the transformers.

    See also
    --------
    himalaya.kernel_ridge.make_column_kernelizer : convenience function for
        combining the outputs of multiple kernelizer objects applied to
        column subsets of the original feature space.
    sklearn.compose.make_column_selector : convenience function for selecting
        columns based on datatype or the columns name with a regex pattern.

    Examples
    --------
    >>> import numpy as np
    >>> from himalaya.kernel_ridge import ColumnKernelizer
    >>> from himalaya.kernel_ridge import Kernelizer
    >>> ck = ColumnKernelizer(
    ...     [("kernel_1", Kernelizer(kernel="linear"), [0, 1, 2]),
    ...      ("kernel_2", Kernelizer(kernel="polynomial"), slice(3, 5))])
    >>> X = np.array([[0., 1., 2., 2., 3.],
                      [0., 2., 0., 0., 3.],
                      [0., 0., 1., 0., 3.],
    ...               [1., 1., 0., 1., 2.]])
    >>> # Kernelize separately the first three columns and the last two
    >>> # columns, creating two kernels of shape (n_samples, n_samples).
    >>> ck.fit_transform(X).shape
    (2, 4, 4)
    """
    # This is not a kernelizer, since it returns multiple kernels
    kernelizer = False

    def __init__(self, transformers, remainder='drop', n_jobs=None,
                 transformer_weights=None, verbose=False, force_cpu=False,
                 force_int_remainder_cols=True):
        self.transformers = transformers
        self.remainder = remainder
        self.sparse_threshold = 0
        self.n_jobs = n_jobs
        self.transformer_weights = transformer_weights
        self.verbose = verbose
        self.force_cpu = force_cpu
        self.force_int_remainder_cols = force_int_remainder_cols

    def _iter(self, fitted=False, *args, **kwargs):
        """
        Generate (name, trans, column, weight) tuples.

        Add a default (linear) Kernelizer to any transformer that does not end
        with a Kernelizer.
        """
        for name, trans, column, weight in super()._iter(
                fitted=fitted, *args, **kwargs):

            if not fitted:
                if trans == 'drop':
                    pass
                elif trans == 'passthrough':
                    trans = Kernelizer()
                elif not _end_with_a_kernel(trans):
                    trans = make_pipeline(trans, Kernelizer())

            yield (name, trans, column, weight)

    def _hstack(self, Xs, **kwargs):
        """Stack the kernels.

        In ColumnTransformer, this methods stacks Xs horizontally.
        Here instead, we stack all kernels in a new dimension.

        Parameters
        ----------
        Ks : array of shape (n_kernels, n_samples, n_samples)
        """
        backend = get_backend()
        return backend.stack(Xs)

    fit = force_cpu_backend(ColumnTransformer.fit)
    fit_transform = force_cpu_backend(ColumnTransformer.fit_transform)
    transform = force_cpu_backend(ColumnTransformer.transform)

    def get_X_fit(self):
        """Helper to get the input data X seen during the fit.

        Returns
        -------
        Xs : list of arrays of shape (n_samples, n_features_i)
            Input arrays for each kernelizer.
        """
        check_is_fitted(self)

        if version.parse(sklearn.__version__) < version.parse("1.4"):
            iter_kwargs = {"replace_strings": True}
        else:
            iter_kwargs = {
                "column_as_labels": False, 
                "skip_drop": True, 
                "skip_empty_columns": True
            }

        Xs = []
        for (_, trans, _, _) in self._iter(fitted=True, **iter_kwargs):
            if hasattr(trans, "get_X_fit"):
                X = trans.get_X_fit()
            else:
                X = trans[-1].get_X_fit()

            Xs.append(X)
        return Xs


def make_column_kernelizer(*transformers, **kwargs):
    """Construct a ColumnKernelizer from the given transformers.

    This is a shorthand for the ColumnKernelizer constructor; it does not
    require, and does not permit, naming the transformers. Instead, they will
    be given names automatically based on their types. It also does not allow
    weighting with ``transformer_weights``.

    Parameters
    ----------
    *transformers : tuples
        Tuples of the form (transformer, columns) specifying the
        transformer objects to be applied to subsets of the data.

        transformer : {'drop', 'passthrough'} or estimator
            Estimator must support ``fit`` and ``transform``.
            Special-cased strings 'drop' and 'passthrough' are accepted as
            well, to indicate to drop the columns or to pass them through
            untransformed, respectively. If the transformer does not return a
            kernel (as informed by the attribute kernelizer=True), a linear
            kernelizer is applied after the transformer.
        columns : str,  array-like of str, int, array-like of int, slice, \
                array-like of bool or callable
            Indexes the data on its second axis. Integers are interpreted as
            positional columns, while strings can reference DataFrame columns
            by name. A scalar string or int should be used where
            ``transformer`` expects X to be a 1d array-like (vector),
            otherwise a 2d array will be passed to the transformer.
            A callable is passed the input data `X` and can return any of the
            above. To select multiple columns by name or dtype, you can use
            :obj:`make_column_selector`.

    remainder : {'drop', 'passthrough'} or estimator, default='drop'
        By default, only the specified columns in `transformers` are
        transformed and combined in the output, and the non-specified
        columns are dropped. (default of ``'drop'``).
        By specifying ``remainder='passthrough'``, all remaining columns that
        were not specified in `transformers` will be automatically passed
        through. This subset of columns is concatenated with the output of
        the transformers.
        By setting ``remainder`` to be an estimator, the remaining
        non-specified columns will use the ``remainder`` estimator. The
        estimator must support ``fit`` and ``transform``.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        n_jobs does not work with with GPU backends.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.

    force_cpu : bool
        If True, computations will be performed on CPU, ignoring the
        current backend. If False, use the current backend.

    Returns
    -------
    column_kernelizer : ColumnKernelizer

    See also
    --------
    himalaya.kernel_ridge.ColumnKernelizer : Class that allows combining the
        outputs of multiple transformer objects used on column subsets
        of the data into a single feature space.

    Examples
    --------
    >>> import numpy as np
    >>> from himalaya.kernel_ridge import make_column_kernelizer
    >>> from himalaya.kernel_ridge import Kernelizer
    >>> ck = make_column_kernelizer(
    ...     (Kernelizer(kernel="linear"), [0, 1, 2]),
    ...     (Kernelizer(kernel="polynomial"), slice(3, 5)))
    >>> X = np.array([[0., 1., 2., 2., 3.],
                      [0., 2., 0., 0., 3.],
                      [0., 0., 1., 0., 3.],
    ...               [1., 1., 0., 1., 2.]])
    >>> # Kernelize separately the first three columns and the last two
    >>> # columns, creating two kernels of shape (n_samples, n_samples).
    >>> ck.fit_transform(X).shape
    (2, 4, 4)
    """
    # transformer_weights keyword is not passed through because the user
    # would need to know the automatically generated names of the transformers
    n_jobs = kwargs.pop('n_jobs', None)
    remainder = kwargs.pop('remainder', 'drop')
    verbose = kwargs.pop('verbose', False)
    force_cpu = kwargs.pop('force_cpu', False)
    if kwargs:
        raise TypeError('Unknown keyword arguments: "{}"'.format(
            list(kwargs.keys())[0]))
    transformer_list = _get_transformer_list(transformers)
    return ColumnKernelizer(transformer_list, n_jobs=n_jobs,
                            remainder=remainder, verbose=verbose,
                            force_cpu=force_cpu)


def _get_transformer_list(estimators):
    """
    Construct (name, trans, column) tuples from list
    """
    transformers, columns = zip(*estimators)
    names, _ = zip(*_name_estimators(transformers))

    transformer_list = list(zip(names, transformers, columns))
    return transformer_list


def _end_with_a_kernel(estimator):
    """Return True if the estimator returns a kernel."""

    if not isinstance(estimator, BaseEstimator):
        raise ValueError("This function requires a scikit-learn estimator.")

    if getattr(estimator, "kernelizer", False):
        return True
    try:
        return _end_with_a_kernel(estimator[-1])
    except TypeError:
        pass

    return False


def _to_cpu(X):
    from ..validation import issparse
    backend = get_backend()
    if issparse(X):
        return X
    else:
        return backend.to_cpu(X)



"""The "torch_cuda" GPU backend, based on PyTorch.

To use this backend, call ``himalaya.backend.set_backend("torch_cuda")``.
"""
from .torch import *  # noqa
import torch

if not torch.cuda.is_available():
    import sys
    if "pytest" in sys.modules:  # if run through pytest
        import pytest
        pytest.skip("PyTorch with CUDA is not available.")
    raise RuntimeError("PyTorch with CUDA is not available.")

from ._utils import _dtype_to_str
from ._utils import warn_if_not_float32

###############################################################################

name = "torch_cuda"


def randn(*args, **kwargs):
    return torch.randn(*args, **kwargs).cuda()


def rand(*args, **kwargs):
    return torch.rand(*args, **kwargs).cuda()


def asarray(x, dtype=None, device="cuda"):
    if dtype is None:
        if isinstance(x, torch.Tensor):
            dtype = x.dtype
        if hasattr(x, "dtype") and hasattr(x.dtype, "name"):
            dtype = x.dtype.name
    if dtype is not None:
        dtype = _dtype_to_str(dtype)
        dtype = getattr(torch, dtype)
    if device is None:
        if isinstance(x, torch.Tensor):
            device = x.device
        else:
            device = "cuda"
    try:
        tensor = torch.as_tensor(x, dtype=dtype, device=device)
    except Exception:
        import numpy as np
        array = np.asarray(x, dtype=_dtype_to_str(dtype))
        tensor = torch.as_tensor(array, dtype=dtype, device=device)
    return tensor


def check_arrays(*all_inputs):
    """Change all inputs into Tensors (or list of Tensors) using the same
    precision and device as the first one. Some tensors can be None.
    """
    all_tensors = []
    all_tensors.append(asarray(all_inputs[0]))
    dtype = all_tensors[0].dtype
    warn_if_not_float32(dtype)
    device = all_tensors[0].device
    for tensor in all_inputs[1:]:
        if tensor is None:
            pass
        elif isinstance(tensor, list):
            tensor = [asarray(tt, dtype=dtype, device=device) for tt in tensor]
        else:
            tensor = asarray(tensor, dtype=dtype, device=device)
        all_tensors.append(tensor)
    return all_tensors


def zeros(shape, dtype="float32", device="cuda"):
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    return torch.zeros(shape, dtype=dtype, device=device)


def to_cpu(array):
    return array.cpu()


def to_gpu(array, device="cuda"):
    return asarray(array, device=device)



from ..backend import get_backend
from ..progress_bar import bar
from ..utils import _batch_or_skip


def predict_weighted_kernel_ridge(Ks, dual_weights, deltas, split=False,
                                  n_targets_batch=None, progress_bar=False,
                                  intercept=None):
    """
    Compute predictions, typically on a test set.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples_test, n_samples_train)
        Test kernels.
    dual_weights : array of shape (n_samples_train, n_targets)
        Dual weights of the kernel ridge model.
    deltas : array of shape (n_kernels, n_targets) or (n_kernels, )
        Log kernel weights for each target.
    split : bool
        If True, the predictions is split across kernels.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.
    intercept : None, or array of shape (n_targets,)
        Intercept added to the predictions. Must be None if split=True.

    Returns
    -------
    Y_hat : array of shape (n_samples_test, n_targets) or \
            (n_kernels, n_samples_test, n_targets) (if split is True)
        Predicted values.
    """
    backend = get_backend()

    Ks, dual_weights, deltas, intercept = backend.check_arrays(
        Ks, dual_weights, deltas, intercept)
    n_samples = Ks.shape[1]
    n_targets = dual_weights.shape[1]
    n_kernels = deltas.shape[0]

    if deltas.ndim == 1:
        deltas = deltas[:, None]

    if split:
        Y_hat_full = backend.zeros_like(
            deltas, shape=(n_kernels, n_samples, n_targets))
    else:
        Y_hat_full = backend.zeros_like(deltas, shape=(n_samples, n_targets))

    if not n_targets_batch:
        n_targets_batch = n_targets

    for start in bar(list(range(0, n_targets, n_targets_batch)),
                     title='predict', use_it=progress_bar):
        batch = slice(start, start + n_targets_batch)
        dual_weights_batch = dual_weights[:, batch]
        deltas_batch = _batch_or_skip(deltas, batch, axis=1)
        chi = backend.matmul(Ks, dual_weights_batch)
        split_predictions = backend.exp(deltas_batch[:, None, :]) * chi
        if split:
            if intercept is not None:
                raise ValueError(
                    "Cannot split the predictions with an intercept.")
            Y_hat_full[:, :, batch] = split_predictions
        else:
            Y_hat_full[:, batch] = split_predictions.sum(0)
            if intercept is not None:
                Y_hat_full[:, batch] += intercept[None, batch]

    return Y_hat_full


def predict_and_score_weighted_kernel_ridge(Ks, dual_weights, deltas, Y,
                                            score_func, split=False,
                                            n_targets_batch=None,
                                            progress_bar=False,
                                            intercept=None):
    """
    Compute predictions, typically on a test set, and compute the score.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples_test, n_samples_train)
        Input kernels.
    dual_weights : array of shape (n_samples_train, n_targets)
        Dual weights of the kernel ridge model.
    deltas : array of shape (n_kernels, n_targets) or (n_kernels, )
        Log kernel weights for each target.
    Y : array of shape (n_samples_test, n_targets)
        Target data.
    score_func : callable
        Function used to compute the score of predictions.
    split : bool
        If True, the predictions is split across kernels.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.
    intercept : None, or array of shape (n_targets,)
        Intercept added to the predictions. To allow split=True, the intercept
        is not added to the predictions but subtracted to the target Y.

    Returns
    -------
    scores : array of shape (n_targets, ) or (n_kernels, n_targets) (if split)
        Prediction score per target.
    """
    backend = get_backend()
    Ks, dual_weights, deltas, Y, intercept = backend.check_arrays(
        Ks, dual_weights, deltas, Y, intercept)

    if deltas.ndim == 1:
        deltas = deltas[:, None]

    n_kernels, _ = deltas.shape
    _, n_targets = Y.shape
    if split:
        scores = backend.zeros_like(Y, shape=(n_kernels, n_targets))
    else:
        scores = backend.zeros_like(Y, shape=(n_targets))

    if n_targets_batch is None:
        n_targets_batch = n_targets
    for start in bar(list(range(0, n_targets, n_targets_batch)),
                     title='predict_and_score', use_it=progress_bar):
        batch = slice(start, start + n_targets_batch)
        predictions = predict_weighted_kernel_ridge(
            Ks, dual_weights[:, batch], _batch_or_skip(deltas, batch, axis=1),
            split=split)
        if intercept is not None:
            score_batch = score_func(Y[:, batch] - intercept[None, batch],
                                     predictions)
        else:
            score_batch = score_func(Y[:, batch], predictions)

        if split:
            scores[:, batch] = score_batch
        else:
            scores[batch] = score_batch

    return scores


def primal_weights_kernel_ridge(dual_weights, X_fit):
    """Compute the primal weights for kernel ridge regression.

    Parameters
    ----------
    dual_weights : array of shape (n_samples_fit, n_targets)
        Dual coefficient of the kernel ridge regression.
    X_fit : array of shape (n_samples_fit, n_features)
        Training features.

    Returns
    -------
    primal_weights : array of shape (n_features, n_targets)
        Primal coefficients of the equivalent ridge regression. The
        coefficients are computed on CPU memory, since they can be large.
    """
    backend = get_backend()
    X_fit = backend.to_cpu(X_fit)
    dual_weights = backend.to_cpu(dual_weights)

    return X_fit.T @ dual_weights


def primal_weights_weighted_kernel_ridge(dual_weights, deltas, Xs_fit):
    """Compute the primal weights for weighted kernel ridge regression.

    Parameters
    ----------
    dual_weights : array of shape (n_samples_fit, n_targets)
        Dual coefficient of the kernel ridge regression.
    deltas : array of shape (n_kernels, n_targets)
        Log of kernel weights.
    Xs_fit : list of arrays of shape (n_samples_fit, n_features)
        Training features. The list should have `n_kernels` elements.

    Returns
    -------
    primal_weights : list of arrays of shape (n_features, n_targets)
        Primal coefficients of the equivalent ridge regression. The
        coefficients are computed on CPU memory, since they can be large.
    """
    backend = get_backend()
    dual_weights = backend.to_cpu(dual_weights)

    primal_weights = []
    for X_fit, deltas_i in zip(Xs_fit, deltas):
        X_fit = backend.to_cpu(X_fit)
        exp_deltas_i = backend.to_cpu(backend.exp(deltas_i))
        primal_weights_i = X_fit.T @ dual_weights * exp_deltas_i[None]
        primal_weights.append(primal_weights_i)

    return primal_weights



import sys
import time


def _format_time(seconds):
    """Format seconds as hh:mm:ss.
    
    Parameters
    ----------
    seconds : float
        Time in seconds.
    
    Returns
    -------
    str
        Formatted time string in hh:mm:ss format.
    """
    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    secs = int(seconds % 60)
    return f"{hours:02d}:{minutes:02d}:{secs:02d}"


def bar(iterable, title='', use_it=True):
    """Simple API for progress_bar.

    Parameters
    ----------
    iterable : iterable
        Iterable that will make the progress bar progress.
    title : str
        Message to include at end of progress bar.
    use_it : bool
        If False, return the iterable unchanged, and does not show a progress
        bar.

    Examples
    --------
    >>> import time
    >>> from himalaya.progress_bar import bar
    >>> for ii in bar(range(10)):
    >>>     time.sleep(0.5)
    """
    if use_it:
        return ProgressBar(title=title, max_value=len(iterable))(iterable)
    else:
        return iterable


class ProgressBar():
    """Generate a command-line progress bar.

    Parameters
    ----------
    max_value : int
        Maximum value of process (e.g. number of samples to process, bytes to
        download, etc.).
    initial_value : int
        Initial value of process, useful when resuming process from a specific
        value, defaults to 0.
    title : str
        Message to include at end of progress bar.
    max_chars : int
        Number of characters to use for progress bar (be sure to save some room
        for the message and % complete as well).
    progress_character : char
        Character in the progress bar that indicates the portion completed.
    spinner : bool
        Show a spinner.  Useful for long-running processes that may not
        increment the progress bar very often.  This provides the user with
        feedback that the progress has not stalled.

    Examples
    --------
    >>> import time
    >>> from himalaya.progress_bar import ProgressBar
    >>> for ii in ProgressBar(title="La barre", max_value=10)(range(10)):
    >>>     time.sleep(0.5)
    """

    spinner_symbols = ['|', '/', '-', '\\']
    template = '\r[{0}{1}] {2:0.0f}% {3} {4:.02f} sec | {5} | {6}'

    def __init__(self, title='', max_value=None, initial_value=0, max_chars=30,
                 progress_character='.', spinner=False, verbose_bool=True):
        self.cur_value = initial_value
        self.max_value = max_value
        self.title = title
        self.max_chars = max_chars
        self.progress_character = progress_character
        self.spinner = spinner
        self.spinner_index = 0
        self.n_spinner = len(self.spinner_symbols)
        self._do_print = verbose_bool
        self.start = time.time()

        self.closed = False
        self.update(initial_value)

    def update(self, cur_value, title=None):
        """Update progressbar with current value of process.

        Parameters
        ----------
        cur_value : number
            Current value of process.  Should be <= max_value (but this is not
            enforced).  The percent of the progressbar will be computed as
            (cur_value / max_value) * 100
        title : str
            Message to display to the right of the progressbar.  If None, the
            last message provided will be used.  To clear the current message,
            pass a null string, ''.
        """
        # Ensure floating-point division so we can get fractions of a percent
        # for the progressbar.
        self.cur_value = cur_value
        max_value = self.max_value or 1
        progress = min(float(self.cur_value) / max_value, 1.)
        num_chars = int(progress * self.max_chars)
        num_left = self.max_chars - num_chars

        # Update the message
        if title is not None:
            self.title = title

        # time from start
        current_time = time.time()
        duration = current_time - self.start

        # Calculate iteration rate and estimated time remaining
        eta_str = ""
        if self.cur_value > 0 and duration > 0:
            iter_per_sec = self.cur_value / duration
            remaining = max_value - self.cur_value
            # Show ETA: 00:00:00 when progress displays as 100% (even if not exactly 1.0)
            # This happens when progress >= 0.995 due to rounding in the display format
            if remaining > 0 and iter_per_sec > 0 and progress < 0.995:
                eta_seconds = remaining / iter_per_sec
                eta_formatted = _format_time(eta_seconds)
                eta_str = f"{iter_per_sec:.2f} it/s, ETA: {eta_formatted}"
            else:
                # Show ETA: 00:00:00 at completion (when progress displays as 100%)
                # to avoid leaving trailing characters
                eta_str = f"{iter_per_sec:.2f} it/s, ETA: 00:00:00"

        # The \r tells the cursor to return to the beginning of the line rather
        # than starting a new line.  This allows us to have a progressbar-style
        # display in the console window.
        bar = self.template.format(self.progress_character * num_chars,
                                   ' ' * num_left, progress * 100,
                                   self.spinner_symbols[self.spinner_index],
                                   duration, self.title, eta_str)
        # Force a flush because sometimes when using bash scripts and pipes,
        # the output is not printed until after the program exits.
        if self._do_print:
            sys.stdout.write(bar)
            sys.stdout.flush()
        # Increment the spinner
        if self.spinner:
            self.spinner_index = (self.spinner_index + 1) % self.n_spinner

        if progress == 1:
            self.close()

    def update_with_increment_value(self, increment_value, title=None):
        """Update progressbar with the value of the increment instead of the
        current value of process as in update().

        Parameters
        ----------
        increment_value : int
            Value of the increment of process.  The percent of the progressbar
            will be computed as
            (self.cur_value + increment_value / max_value) * 100
        title : str
            Message to display to the right of the progressbar.  If None, the
            last message provided will be used.  To clear the current message,
            pass a null string, ''.
        """
        self.cur_value += increment_value
        self.update(self.cur_value, title)

    def close(self):
        """Close the progress bar."""
        if not self.closed:
            sys.stdout.write('\n')
            sys.stdout.flush()
            self.closed = True

    def __call__(self, sequence):
        sequence = iter(sequence)
        while True:
            try:
                yield next(sequence)
                self.update_with_increment_value(1)
            except StopIteration:
                return



import re
from pathlib import Path
from setuptools import find_packages, setup

# get version from himalaya/__init__.py
__version__ = 0.0
with open("himalaya/__init__.py") as f:
    infos = f.readlines()
for line in infos:
    if "__version__" in line:
        match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", line)
        __version__ = match.groups()[0]

# read the contents of the README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

requirements = [
    "numpy",
    "scikit-learn",
    # "cupy",  # optional backend
    # "torch",  # optional backend, 1.9+ preferred
    # "matplotlib",  # for visualization only
    # "pytest",  # for testing only
]

extras_require = {
    "all_backends": ["cupy", "torch"],
    "viz": ["matplotlib"],
    "test": ["pytest", "matplotlib", "cupy", "torch"],
    "github": ["pytest", "matplotlib", "torch", "pytest-rerunfailures"],
}

extras_require["all"] = sum(list(extras_require.values()), [])
extras_require["doc"] = [
    "numpydoc",
    "sphinx",
    "sphinx_gallery",
    "sphinxcontrib-mermaid",
]
extras_require["doc"] += extras_require["viz"] + extras_require["all_backends"]

if __name__ == "__main__":
    setup(
        name="himalaya",
        maintainer="Tom Dupre la Tour",
        maintainer_email="tomdlt@berkeley.edu",
        description="Multiple-target machine learning",
        license="BSD (3-clause)",
        version=__version__,
        packages=find_packages(),
        url="https://github.com/gallantlab/himalaya",
        install_requires=requirements,
        extras_require=extras_require,
        long_description=long_description,
        long_description_content_type="text/x-rst",
    )



import types
import importlib
import warnings
from functools import wraps

ALL_BACKENDS = [
    "numpy",
    "cupy",
    "torch",
    "torch_cuda",
]

CURRENT_BACKEND = "numpy"

MATCHING_CPU_BACKEND = {
    "numpy": "numpy",
    "cupy": "numpy",
    "torch": "torch",
    "torch_cuda": "torch",
}


def set_backend(backend, on_error="raise"):
    """Set the backend using a global variable, and return the backend module.

    Parameters
    ----------
    backend : str or module
        Name or module of the backend.
    on_error : str in {"raise", "warn"}
        Define what is done if the backend fails to be loaded.
        If "warn", this function only warns, and keeps the previous backend.
        If "raise", this function raises on errors.

    Returns
    -------
    module : python module
        Module of the backend.
    """
    global CURRENT_BACKEND

    try:
        if isinstance(backend, types.ModuleType):  # get name from module
            backend = backend.name

        if backend not in ALL_BACKENDS:
            raise ValueError("Unknown backend=%r" % (backend, ))

        module = importlib.import_module(__package__ + "." + backend)
        CURRENT_BACKEND = backend
    except Exception as error:
        if on_error == "raise":
            raise error
        elif on_error == "warn":
            warnings.warn(f"Setting backend to {backend} failed: {str(error)}."
                          f"Falling back to {CURRENT_BACKEND} backend.")
            module = get_backend()
        else:
            raise ValueError('Unknown value on_error=%r' % (on_error, ))

    return module


def get_backend():
    """Get the current backend module.

    Returns
    -------
    module : python module
        Module of the backend.
    """
    module = importlib.import_module(__package__ + "." + CURRENT_BACKEND)
    return module


def _dtype_to_str(dtype):
    """Cast dtype to string, such as "float32", or "float64"."""
    if isinstance(dtype, str):
        return dtype
    elif hasattr(dtype, "name"):  # works for numpy and cupy
        return dtype.name
    elif "torch." in str(dtype):  # works for torch
        return str(dtype)[6:]
    elif dtype is None:
        return None
    else:
        raise NotImplementedError()


def force_cpu_backend(func):
    """Decorator to force the use of a CPU backend."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        # skip if the object does not force cpu use
        if not hasattr(args[0], "force_cpu") or not args[0].force_cpu:
            return func(*args, **kwargs)

        # set corresponding cpu backend
        original_backend = get_backend().name
        temp_backend = MATCHING_CPU_BACKEND[original_backend]
        set_backend(temp_backend)

        # run function
        result = func(*args, **kwargs)

        # set back original backend
        set_backend(original_backend)
        return result

    return wrapper


def _add_error_message(func, msg=""):
    """Decorator to add a custom error message to a function."""

    @wraps(func)
    def with_error_message(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise RuntimeError(
                f"{msg}\nOriginal error:\n{type(e).__name__}: {e}")

    return with_error_message


_already_warned = [False]


def warn_if_not_float32(dtype):
    """Warn if X is not float32."""
    if _already_warned[0]:  # avoid warning multiple times
        return None

    if _dtype_to_str(dtype) != "float32":
        backend = get_backend()
        warnings.warn(
            f"GPU backend {backend.name} is much faster with single "
            f"precision floats (float32), got input in {dtype}. "
            "Consider casting your data to float32.", UserWarning)
        _already_warned[0] = True



from ._group_lasso import solve_sparse_group_lasso
from ._group_lasso import solve_sparse_group_lasso_cv
from ._sklearn_api import SparseGroupLassoCV

__all__ = [
    "solve_sparse_group_lasso",
    "solve_sparse_group_lasso_cv",
    "SparseGroupLassoCV",
]
